from django.db import models


class Order(models.Model):
    NEW = 'NEW'
    DECLINED = 'DECLINED'
    LOAN_CREATED = 'LOAN_CREATED'
    EXPIRED = 'EXPIRED'
    STATUS_CHOICES = {
        NEW: 'Новая',
        LOAN_CREATED: 'Займ создан',
        DECLINED: 'Отклонена',
        EXPIRED: 'Истекла',
    }
    created = models.DateTimeField('Создана', auto_now_add=True)
    updated = models.DateTimeField('Изменена', auto_now=True)
    user_id = models.PositiveIntegerField()
    amount_requested = models.DecimalField('Запрошенная сумма', max_digits=7, decimal_places=0)
    amount_approved = models.DecimalField('Одобренная сумма', max_digits=7, decimal_places=0, null=True)
    amount_approved_max = models.DecimalField('Макс. одобренная сумма', max_digits=7, decimal_places=0, null=True)
    period_requested = models.PositiveSmallIntegerField('Запрошенный период')
    period_approved = models.PositiveSmallIntegerField('Одобренный период', blank=True, null=True)
    period_free = models.PositiveSmallIntegerField('Беспроцентный период')
    interests = models.DecimalField('Процентная ставка', max_digits=3, decimal_places=2)
    sign = models.PositiveSmallIntegerField('Подпись')
    attempts = models.PositiveSmallIntegerField('Попытки', default=0, help_text='Попытки ввода подписи')
    status = models.CharField('Статус', max_length=30, choices=sorted(STATUS_CHOICES.items()), default=NEW)
    promocode_code = models.CharField('Промокод', max_length=16, blank=True)
    contract_demo = models.FileField('Демо договора', upload_to="files/", blank=True)
    is_first_loan = models.BooleanField('Первый займ', null=True, blank=True,
                                        help_text='True - ранее у юзера не было займов')
    is_first_order = models.BooleanField('Первая заявка', null=True, blank=True,
                                         help_text='True - ранее у юзера не было заявок')
    comment = models.CharField('Комментарий', max_length=1024, blank=True)

    def save(self, *args, **kwargs):
        pass

    def __str__(self):
        return f'{self.id}'

    class Meta:
        managed = False
        verbose_name = 'Заявка'
        verbose_name_plural = 'Заявки'



# Copyright (c) 2017-2022, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.
from django.forms import ModelForm

from .renderers import (
    Column,
    FieldSet,
    Layout,
    FormLayout,
    LayoutNode,
    Row,
    FormSet,
    Span,
    Caption,
)


class FormAjaxCompleteMixin:
    pass


class FormDependentSelectMixin:
    pass


__all__ = (
    "Caption",
    "Column",
    "Layout",
    "LayoutNode",
    "Row",
    "Span",
    "FieldSet",
    "FormLayout",
    "FormSet",
    "ModelForm",
    "FormAjaxCompleteMixin",
    "FormDependentSelectMixin",
)



import json
import re
from django.http import HttpResponse, JsonResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST
from django.views.generic import TemplateView


def _extract_urls(data):
    """
    Given html string of <script src=""> tags, extracts list of static files stripping version info

    >>> _extract_urls(
    >>>     '<script src="/_dash-component-suites/dash/deps/polyfill@7.v2_0_0m1633432698.12.1.min.js"></script>\n<script src="/_dash-component-suites/dash/deps/react@16.v2_0_0m1633432698.14.0.min.js">'  # noqa
    >>> )
    /viewflow/js/contrib/dash/deps/polyfill@7.12.1.min.js
    """
    for url in re.findall('src=\"(.*)\"', data):
        _, url = url.split('/_dash-component-suites/')
        start, end = url.rsplit('.v')
        extention = end.split('.', 1)[1]
        yield f"viewflow/js/contrib/{start}.{extention}"


class DashboardView(TemplateView):
    template_name = 'viewflow/contrib/plotly.html'
    viewset = None

    def get_context_data(self, **kwargs):
        return super().get_context_data(
            dash_scripts_urls=_extract_urls(self.viewset.dash_app._generate_scripts_html()),
            dash_config_html=self.viewset.dash_app._generate_config_html()
        )


def layout_endpoint(request, viewset):
    dash_response = viewset.dash_app.serve_layout()
    return HttpResponse(dash_response.data, content_type='application/json')


def dependencies_endpoint(request, viewset):
    return JsonResponse(viewset.dash_app._callback_list, safe=False)


@csrf_exempt
@require_POST
def update_component_endpoint(request, viewset):
    try:
        request_body = json.loads(request.body.decode())

        output = request_body["output"]
        outputs_list = request_body["outputs"]

        inputs = request_body.get("inputs", [])
        state = request_body.get("state", [])

        callback_data = viewset.dash_app.callback_map[output]

    except (json.JSONDecodeError, KeyError) as e:
        return JsonResponse({'code': 400, 'message': f'Request body error: {e}'}, status=400)
    else:
        callback_args = [
            [item.get("value") for item in items]
            if isinstance(items, list) else items.get("value")
            for items in inputs + state
        ]

        callback_func = callback_data["callback"]
        result = callback_func(*callback_args, outputs_list=outputs_list)
        return HttpResponse(result, content_type='application/json')



from ..base import Flow
from .mixins import NodeDetailMixin, NodeExecuteMixin
from .nodes import (
    End,
    Function,
    Handle,
    If,
    Join,
    NSubprocess,
    Obsolete,
    Split,
    SplitFirst,
    Start,
    StartHandle,
    StartSubprocess,
    Subprocess,
    Switch,
    View,
)
from .viewset import FlowViewset, FlowAppViewset, WorkflowAppViewset


__all__ = (
    "End",
    "Flow",
    "FlowAppViewset",
    "FlowViewset",
    "Function",
    "Handle",
    "If",
    "Join",
    "NodeDetailMixin",
    "NodeExecuteMixin",
    "NSubprocess",
    "Obsolete",
    "Split",
    "SplitFirst",
    "Start",
    "StartHandle",
    "StartSubprocess",
    "Subprocess",
    "Switch",
    "View",
    "WorkflowAppViewset",
)



from .status import STATUS


class Act(object):
    """Shortcut to access activation data."""

    @property
    def process(self):
        """Shortcut for lambda activation: activation.process...)"""

        class Lookup(object):
            def __getattribute__(self, name):
                return lambda activation: getattr(activation.process, name)

        return Lookup()

    @property
    def task(self):
        """Shortcut for lambda activation: activation.task...)"""

        class Lookup(object):
            def __getattribute__(self, name):
                return lambda activation: getattr(activation.task, name)

        return Lookup()


act = Act()


def get_next_process_task(manager, process, user):
    task = manager.filter(process=process, owner=user, status=STATUS.ASSIGNED).first()

    # lookup for a task in a queue
    if task is None:
        task = (
            manager.user_queue(user).filter(process=process, status=STATUS.NEW).first()
        )

    # lookup for a job
    if task is None:
        task = manager.filter(
            process=process,
            flow_task_type="JOB",
            status__in=[STATUS.NEW, STATUS.SCHEDULED, STATUS.STARTED],
        ).first()

    return task



# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.

import json
import re
from copy import copy
from typing import List, Type, Union
from functools import lru_cache
from xml.etree import ElementTree

from django import forms
from django.core.exceptions import ValidationError
from django.core.serializers.json import DjangoJSONEncoder
from django.utils.encoding import force_str
from typing import Optional

from django.utils.formats import get_format
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _

from viewflow.utils import MARKER


AUTO = MARKER("AUTO")


class WidgetRenderer(object):
    tag = "vf-field-input"

    def __init__(
        self,
        root: ElementTree.Element,
        bound_field: forms.BoundField,
        layout_node: Optional["LayoutNode"] = None,
    ):
        self.root = root
        self.bound_field = bound_field
        self.layout_node = layout_node

    def format_value(self, value):
        if isinstance(value, (tuple, list)):
            return ",".join(f"{item}" for item in value)
        return f"{value}"

    def create_root(self, context):
        attrs = {
            key: str(value) if value is not True else key
            for key, value in context["widget"]["attrs"].items()
            if value is not False
        }
        attrs["name"] = context["widget"]["name"]
        value = context["widget"]["value"]
        if value is not None:
            attrs["value"] = self.format_value(value)
        if self.bound_field.label:
            attrs["label"] = self.bound_field.label
        if self.bound_field.errors:
            attrs["error"] = self.bound_field.errors[0]
        if self.bound_field.help_text:
            attrs["help-text"] = self.bound_field.help_text
        return ElementTree.SubElement(self.root, self.tag, **attrs)

    def render(self, template_name, context, request=None):
        return self.create_root(context)

    @staticmethod
    @lru_cache(maxsize=None)
    def get_renderer(widget: forms.Widget) -> "Type[WidgetRenderer]":
        from viewflow.conf import settings

        for widget_class in type(widget).mro()[:-2]:
            try:
                return settings.WIDGET_RENDERERS[widget_class]
            except KeyError:
                continue

        return InputRenderer


class CheckboxRenderer(WidgetRenderer):
    tag = "vf-field-checkbox"


class InputRenderer(WidgetRenderer):
    def create_root(self, context):
        root = super().create_root(context)
        root.attrib["type"] = context["widget"]["type"]
        return root


class FileInputRenderer(InputRenderer):
    tag = "vf-field-file"


class HiddenInputRenderer(WidgetRenderer):
    tag = "input"

    def create_root(self, context):
        root = super().create_root(context)
        root.attrib["type"] = "hidden"
        return root


class MultipleHiddenInputRenderer(WidgetRenderer):
    tag = "input"

    def create_root(self, context):
        parent = self.root
        self.root = ElementTree.SubElement(self.root, "div")

        for subwidget in context["widget"]["subwidgets"]:
            child = super().create_root({"widget": subwidget})
            child.attrib["type"] = "hidden"

        root, self.root = self.root, parent
        return root


class PasswordInputRenderer(InputRenderer):
    tag = "vf-field-password"


class TextareaRenderer(WidgetRenderer):
    tag = "vf-field-textarea"


class TrixEditorRenderer(TextareaRenderer):
    tag = "vf-field-editor"


class InlineCalendarRenderer(InputRenderer):
    tag = "vf-field-inline-calendar"

    def find_target_date_format(self, proposed_formats):
        if not proposed_formats:
            formats = get_format("DATE_INPUT_FORMATS")
        else:
            formats = proposed_formats

        supported_formats = [  # input mask friendly formats
            r"%Y-%m-%d",
            r"%m-%d-%Y",
            r"%Y-%d-%m",
            r"%d-%m-%Y",
        ]
        for date_format in formats:
            lookup = date_format.replace(".", "-").replace("/", "-")
            if lookup in supported_formats:
                separator = date_format[2]
                return lookup.replace("-", separator)

        raise ValueError(
            "Widget should accept one of input-mask friendly date input format."
        )

    def create_root(self, context):
        root = super().create_root(context)

        widget_format = self.bound_field.field.widget.format
        root.attrib["format"] = self.find_target_date_format(
            [widget_format] if widget_format else self.bound_field.field.input_formats
        )
        return root


class DateInputRenderer(InlineCalendarRenderer):
    tag = "vf-field-date"

    def date_format_placeholder(self, date_format):
        i18n = {
            r"%Y": _("YYYY"),
            r"%m": _("MM"),
            r"%d": _("DD"),
        }
        split_format = re.split(r"([.\-/])", date_format)
        items = [i18n.get(item, item) for item in split_format]
        return format_lazy("{}" * len(items), *items)

    def create_root(self, context):
        root = super().create_root(context)
        root.attrib["placeholder"] = self.date_format_placeholder(root.attrib["format"])
        return root


class DateTimeInputRenderer(InlineCalendarRenderer):
    tag = "vf-field-date"

    def find_target_date_format(self, proposed_formats):
        if not proposed_formats:
            formats = get_format("DATE_INPUT_FORMATS")
        else:
            formats = proposed_formats

        supported_formats = [  # input mask friendly formats
            r"%Y-%m-%d %H:%M",
            r"%m-%d-%Y %H:%M",
            r"%Y-%d-%m %H:%M",
            r"%d-%m-%Y %H:%M",
            r"%Y-%m-%d %H:%M:%S",
            r"%m-%d-%Y %H:%M:%S",
            r"%Y-%d-%m %H:%M:%S",
            r"%d-%m-%Y %H:%M:%S",
            r"%b-%d-%Y %I:%M %p",
        ]
        for date_format in formats:
            lookup = date_format.replace(".", "-").replace("/", "-")
            if lookup in supported_formats:
                separator = date_format[2]
                return lookup.replace("-", separator)

        raise ValueError(
            "Widget should accept one of input-mask friendly date input format."
        )

    def date_format_placeholder(self, date_format):
        i18n = {
            r"%Y": _("YYYY"),
            r"%m": _("MM"),
            r"%d": _("DD"),
            r"%H": _("HH"),
            r"%M": _("MM"),
            r"%S": _("SS"),
        }
        split_format = re.split(r"([.\-/\:\s])", date_format)
        items = [i18n.get(item, item) for item in split_format]
        return format_lazy("{}" * len(items), *items)

    def create_root(self, context):
        root = super().create_root(context)
        root.attrib["placeholder"] = self.date_format_placeholder(root.attrib["format"])
        return root


class FormDataEncoder(DjangoJSONEncoder):
    def default(self, obj):
        try:
            return super().default(self, obj)
        except TypeError:
            return str(obj)


class SelectRenderer(WidgetRenderer):
    tag = "vf-field-select"

    def create_root(self, context):
        root = super().create_root(context)
        root.attrib["optgroups"] = json.dumps(
            [
                {
                    "name": name or "",
                    "options": {
                        key: value
                        for key, value in options[0].items()
                        if key not in ["template_name", "type", "wrap_label"]
                    },
                }
                for name, options, _ in context["widget"]["optgroups"]
            ],
            cls=FormDataEncoder,
        )
        return root


class SelectMultipleRenderer(SelectRenderer):
    tag = "vf-field-select-multiple"


class CheckboxSelectMultipleRenderer(SelectRenderer):
    tag = "vf-field-checkbox-select"


class DependentSelectRenderer(SelectRenderer):
    tag = "vf-field-select-dependent"


class TimeInputRenderer(InputRenderer):
    tag = "vf-field-time"

    def time_format_placeholder(self, time_format):
        i18n = {
            r"%H": _("HH"),
            r"%M": _("MM"),
            r"%S": _("SS"),
            r"%I": _("HH"),
            r"%p": _("AM/PM"),
        }
        split_format = re.split(r"([.\-/\:\s])", time_format)
        items = [i18n.get(item, item) for item in split_format]
        return format_lazy("{}" * len(items), *items)

    def create_root(self, context):
        root = super().create_root(context)

        widget_format = self.bound_field.field.widget.format
        input_format = (
            widget_format if widget_format else self.bound_field.field.input_formats[0]
        )
        root.attrib["format"] = input_format
        root.attrib["placeholder"] = self.time_format_placeholder(input_format)
        return root


class RadioSelectRenderer(SelectRenderer):
    tag = "vf-field-radio-select"


class FormWidgetRenderer(WidgetRenderer):
    tag = "vf-field-form"

    def render(self, template_name, context, request=None):
        root = self.create_root(context)

        wrapper = ElementTree.SubElement(
            root,
            "div",
            {
                "class": "vf-form__formset",
            },
        )
        title = ElementTree.SubElement(
            wrapper,
            "h3",
            {
                "class": "mdc-typography--subheading2 vf-form__formset-header",
            },
        )
        title.text = force_str(self.bound_field.label)

        form = context["widget"]["value"]
        if hasattr(form, "layout"):
            wrapper.append(form.layout.render_form(form))
        else:
            wrapper.append(FormLayout().render_form(form))
        return root


class FormSetWidgetRenderer(WidgetRenderer):
    tag = "vf-field-formset"

    def render_form(self, root: ElementTree.Element, form):
        if isinstance(self.layout_node, FormSet):
            # put the form inside a card
            grid = ElementTree.SubElement(
                root,
                "div",
                {
                    "class": "mdc-layout-grid__cell"
                    f" mdc-layout-grid__cell--span-{self.layout_node.card_desktop}-desktop"
                    f" mdc-layout-grid__cell--span-{self.layout_node.card_tablet}-tablet"
                    f" mdc-layout-grid__cell--span-{self.layout_node.card_mobile}-mobile"
                },
            )
            root = ElementTree.SubElement(
                grid,
                "div",
                {
                    "class": "mdc-card",
                },
            )

        if hasattr(form, "layout"):
            layout = copy(form.layout)
            if "DELETE" in form.fields and "DELETE" not in layout.children:
                layout.children = layout.children[:]
                layout.children.append("DELETE")

            wrapper = ElementTree.SubElement(
                root,
                "div",
                {
                    "class": "vf-form__formset-form",
                },
            )
            wrapper.append(layout.render_form(form))
        else:
            if "DELETE" in form.fields:
                form_layout = Layout(
                    Row(
                        Column(
                            Row(
                                *[
                                    field.name
                                    for field in form.visible_fields()
                                    if field.name != "DELETE"
                                ]
                            ),
                            desktop=10,
                        ),
                        Column("DELETE", desktop=2),
                    )
                )  # if not isinstance(self.layout_node, FormSet) else FormLayout()
            else:
                form_layout = form_layout = Layout(
                    Row(*[field.name for field in form.visible_fields()]),
                )  # if not isinstance(self.layout_node, FormSet) else FormLayout()
            root.append(form_layout.render_form(form))

    def append_form_template(self, root: ElementTree.Element, formset):
        wrapper = ElementTree.SubElement(
            root,
            "script",
            {
                "type": "form-template",
            },
        )
        self.render_form(wrapper, formset.empty_form)

    def render(self, template_name, context, request=None):
        formset = context["widget"]["value"]

        # do not pass whole formset as the attrubute value
        context["widget"]["value"] = None

        root = self.create_root(context)

        wrapper = ElementTree.SubElement(
            root,
            "div",
            {
                "class": "vf-form__formset",
            },
        )
        title = ElementTree.SubElement(
            wrapper,
            "h3",
            {
                "class": "mdc-typography--subheading2 vf-form__formset-header",
            },
        )
        title.text = force_str(self.bound_field.label)

        self.append_form_template(wrapper, formset)

        forms_wrapper_class = "vf-form__formset-forms"
        if isinstance(self.layout_node, FormSet):
            forms_wrapper_class = "mdc-layout-grid__inner vf-form__formset-forms"
        forms_wrapper = ElementTree.SubElement(
            wrapper,
            "div",
            {
                "class": forms_wrapper_class,
            },
        )

        for form in formset:
            self.render_form(forms_wrapper, form)

        wrapper.append(FormLayout().render_form(formset.management_form))

        button = ElementTree.SubElement(
            wrapper,
            "button",
            {
                "class": "mdc-button mdc-button--outlined vf-from__formset-button",
                "style": "width: 100%;margin-bottom: 16px",
            },
        )
        ElementTree.SubElement(button, "i", {"class": "material-icons"}).text = "add"

        return root


class AjaxModelSelectRenderer(InputRenderer):
    tag = "vf-field-autocomplete"

    def create_root(self, context):
        root = super().create_root(context)

        field = self.bound_field.field
        value_label = ""
        try:
            obj = field.to_python(self.bound_field.value())
            if obj is not None:
                value_label = field.label_from_instance(obj)
            if value_label is None:
                value_label = ""
        except ValidationError:
            pass

        root.attrib["type"] = "hidden"
        root.attrib["value-label"] = value_label
        return root


class AjaxMultipleModelSelectRenderer(InputRenderer):
    tag = "vf-field-autocomplete-multi"

    def create_root(self, context):
        root = super().create_root(context)

        field = self.bound_field.field
        initial_values = [
            {
                'value': field.label_from_instance(item),
                'data': {'id': field.prepare_value(item)},
            }
            for item in field.to_python(self.bound_field.value())
        ]
        root.attrib["initial"] = json.dumps(initial_values)
        return root


class FormLayout:
    """Default form layout."""

    def append_non_field_errors(self, form: forms.Form, root: ElementTree.Element):
        errors = form.non_field_errors()
        errors.extend(
            form.error_class(bound_field.errors)
            for bound_field in form.hidden_fields()
            if bound_field.errors
        )

        if errors:
            wrapper = ElementTree.SubElement(root, "div", {"class": "vf-form__errors"})
            for error in errors:
                child = ElementTree.SubElement(
                    wrapper,
                    "div",
                    {
                        "class": "vf-form__error",
                    },
                )
                child.text = str(error)

    def append_hidden_fields(self, form: forms.Form, root: ElementTree.Element):
        hidden_fields = form.hidden_fields()
        if hidden_fields:
            wrapper = ElementTree.SubElement(
                root, "div", {"class": "vf-form__hiddenfields"}
            )
            for bound_field in hidden_fields:
                self.append_field(wrapper, bound_field)

    def append_visible_fields(self, form: forms.Form, root: ElementTree.Element):
        visible_fields = form.visible_fields()
        if visible_fields:
            wrapper = ElementTree.SubElement(
                root, "div", {"class": "vf-form__visiblefields mdc-layout-grid__inner"}
            )
            for bound_field in form.visible_fields():
                container = ElementTree.SubElement(
                    wrapper,
                    "div",
                    {"class": "mdc-layout-grid__cell mdc-layout-grid__cell--span-12"},
                )
                self.append_field(container, bound_field)

    def append_field(
        self,
        root: ElementTree.Element,
        bound_field: forms.BoundField,
        renderer_class=None,
        layout_node=None,
    ):
        if renderer_class is None:
            renderer_class = WidgetRenderer.get_renderer(bound_field.field.widget)

        widget = bound_field.field.widget
        if bound_field.field.localize:
            widget.is_localized = True
        attrs = bound_field.build_widget_attrs({}, widget)
        if bound_field.auto_id and "id" not in widget.attrs:
            attrs.setdefault("id", bound_field.auto_id)

        widget.render(
            name=bound_field.html_name,
            value=bound_field.value(),
            attrs=attrs,
            renderer=renderer_class(root, bound_field, layout_node),
        )

    def render_form(self, form: forms.Form) -> ElementTree.Element:
        root = ElementTree.Element("div", {"class": "vf-form mdc-layout-grid"})
        self.append_non_field_errors(form, root)
        self.append_hidden_fields(form, root)
        self.append_visible_fields(form, root)
        return root

    def render(self, form: forms.Form, default_layout: "FormLayout" = None):
        return ElementTree.tostring(
            self.render_form(form),
            encoding="unicode",
            method="html",
        )


class Layout(FormLayout):
    """Customizable form layout."""

    def __init__(self, *elements, **kwargs):
        self.children = _convert_to_children(elements)
        super().__init__(**kwargs)

    def append_visible_fields(self, form: forms.Form, root: ElementTree.Element):
        wrapper = ElementTree.SubElement(
            root, "div", {"class": "vf-form__visiblefields mdc-layout-grid__inner"}
        )

        Column(*self.children).append(self, form, wrapper)


class LayoutNode(object):
    """Base class for self-rendered nodes."""

    def __init__(self, desktop=AUTO, tablet=AUTO, mobile=AUTO):
        assert desktop == AUTO or 1 <= desktop <= 12
        self.desktop = desktop

        assert tablet == AUTO or 1 <= tablet <= 8
        self.tablet = tablet

        assert mobile == AUTO or 1 <= mobile <= 4
        self.mobile = mobile

    def append(self, layout: FormLayout, form: forms.Form, root: ElementTree.Element):
        raise NotImplementedError("Subclass should override this")


class Column(LayoutNode):
    """Place elements vertically stacked, one under another.

    Example:
        layout = Layout(
            Row(
                Column('first_name', 'last_name', desktop=8, tablet=6)
                'sex_options'
            )
        )
    """

    def __init__(self, *elements, **kwargs):
        self.id_ = kwargs.pop("id_", None)
        self.children = _convert_to_children(elements)
        super().__init__(**kwargs)

    def append(self, layout: FormLayout, form: forms.Form, root: ElementTree.Element):
        if self.children:
            wrapper = ElementTree.SubElement(
                root,
                "div",
                {
                    "class": "vf-form-column mdc-layout-grid__cell mdc-layout-grid__cell--span-12"
                },
            )
            if self.id_:
                wrapper.attrib["id"] = self.id_
            for child in self.children:
                child.append(layout, form, wrapper)


class Row(LayoutNode):
    """Spread elements over a single line.

    Example:

        layout = Layout(
            Row(
                'first_name',
                Row('last_name', 'sex', tablet=5)
            )
        )
    """

    def __init__(self, *elements, **kwargs):
        self.id_ = kwargs.pop("id_", None)
        self.children = _convert_to_children(elements)
        super().__init__(**kwargs)

    def append(self, layout: FormLayout, form: forms.Form, root: ElementTree.Element):
        desktop = _children_sizes(
            [child.desktop for child in self.children],
            grid_size=12,
            grid_name="desktop",
            keep_in_row=True,
        )
        tablet = _children_sizes(
            [child.tablet for child in self.children],
            grid_size=8,
            grid_name="tablet",
            keep_in_row=False,
        )
        mobile = _children_sizes(
            [child.mobile for child in self.children],
            grid_size=4,
            grid_name="mobile",
            keep_in_row=False,
        )

        def append_child(wrapper, child, desktop: int, tablet: int, mobile: int):
            classes = " ".join(
                [
                    "mdc-layout-grid__cell",
                    f"mdc-layout-grid__cell--span-{desktop}-desktop",
                    f"mdc-layout-grid__cell--span-{tablet}-tablet",
                    f"mdc-layout-grid__cell--span-{mobile}-mobile",
                ]
            )

            element = ElementTree.SubElement(
                wrapper,
                "div",
                {
                    "class": classes,
                },
            )
            child.append(layout, form, element)

        wrapper = ElementTree.SubElement(
            root,
            "div",
            {
                "class": "vf-form-row mdc-layout-grid__inner",
            },
        )
        if self.id_:
            wrapper.attrib["id"] = self.id_
        for child_data in zip(self.children, desktop, tablet, mobile):
            append_child(wrapper, *child_data)


class Span(LayoutNode):
    """Span a form field over several columns.

    Example::
        layout = Layout(
            Row(Span('first_name'), Span('last_name'))
            Row(
                Span('email', tablet=6, mobile=3),
                'sex'
            )
        )

    By default span is auto-sized. On a desktop all auto-sized elements
    would be spread equally over the free place of a row, non occupied by
    elements with specific sizes.

    On mobile and tablet if all elements in a row have auto-sizes,
    each element would be placed in a new line. If even one element
    in a row has a specific size, all auto-sized elements would be
    kept in a single line, like on a desktop.

    """

    def __init__(self, field_name, **kwargs):
        self.field_name = field_name
        super().__init__(**kwargs)

    def __str__(self):
        return f"Field {self.field_name} <{self.desktop}, {self.tablet}, {self.mobile}>"

    def append(self, layout: FormLayout, form: forms.Form, root: ElementTree.Element):
        try:
            bound_field = form[self.field_name]
        except KeyError as exc:
            raise ValueError(
                f"{self.field_name} field not found in the {type(form).__name__}"
            ) from exc

        layout.append_field(root, bound_field, layout_node=self)


class Caption(LayoutNode):
    def __init__(self, text, **kwargs):
        self.text = text
        super().__init__(*kwargs)

    def append(self, layout: FormLayout, form: forms.Form, root: ElementTree.Element):
        wrapper = ElementTree.SubElement(
            root,
            "div",
            {
                "class": "vf-form-column mdc-layout-grid__cell mdc-layout-grid__cell--span-12"
            },
        )
        ElementTree.SubElement(
            wrapper,
            "h6",
            {
                "class": "mdc-typography--caption",
                "style": "margin-bottom: 16px;margin-top: 16px;",
            },
        ).text = self.text


class FieldSet(Column):
    def __init__(self, title, *elements, **kwargs):
        self.title = title
        super().__init__(*elements, **kwargs)

    def append(self, layout: FormLayout, form: forms.Form, root: ElementTree.Element):
        wrapper = ElementTree.SubElement(
            root,
            "div",
            {
                "class": "vf-form__formset",
            },
        )
        title = ElementTree.SubElement(
            wrapper,
            "h3",
            {
                "class": "mdc-typography--subheading2 vf-form__formset-header",
            },
        )
        title.text = force_str(self.title)
        super().append(layout, form, wrapper)


class FormSet(Span):
    """Render stacked inline."""

    def __init__(
        self,
        formset_field_name: str,
        card_desktop: int = 12,
        card_tablet: int = 8,
        card_mobile: int = 4,
        **kwargs,
    ):
        assert 1 <= card_desktop <= 12
        self.card_desktop = card_desktop

        assert 1 <= card_tablet <= 8
        self.card_tablet = card_tablet

        assert 1 <= card_mobile <= 4
        self.card_mobile = card_mobile

        super().__init__(formset_field_name, **kwargs)


def _convert_to_children(elements: List[Union[LayoutNode, str]]):
    result = []
    for element in elements:
        if isinstance(element, LayoutNode):
            result.append(element)
        elif isinstance(element, str):
            result.append(Span(element))
        else:
            raise ValueError(f"Unknown element {element} type {type(element)}")
    return result


def _children_sizes(spans, grid_size=12, grid_name="desktop", keep_in_row=True):
    bound = sum(span for span in spans if span != AUTO)
    auto_count = sum(1 for span in spans if span == AUTO)

    if bound == 0 and not keep_in_row:
        # If children AUTO-sized - put every child on the own row
        return [grid_size for _ in spans]
    else:
        rest = grid_size - bound
        if rest < 0 or (auto_count != 0 and grid_size % auto_count) != 0:
            raise ValueError(
                f"Can't equally spread {spans} over {grid_size} columns on a {grid_name} grid"
            )
        return [rest // auto_count if child == AUTO else child for child in spans]


WIDGET_RENDERERS = {
    forms.CheckboxInput: CheckboxRenderer,
    forms.CheckboxSelectMultiple: CheckboxSelectMultipleRenderer,
    forms.DateInput: DateInputRenderer,
    forms.DateTimeInput: DateTimeInputRenderer,
    forms.FileInput: FileInputRenderer,
    forms.HiddenInput: HiddenInputRenderer,
    forms.MultipleHiddenInput: MultipleHiddenInputRenderer,
    forms.PasswordInput: PasswordInputRenderer,
    forms.RadioSelect: RadioSelectRenderer,
    forms.Select: SelectRenderer,
    forms.SelectMultiple: SelectMultipleRenderer,
    forms.Textarea: TextareaRenderer,
    forms.TimeInput: TimeInputRenderer,
}



from django.contrib import admin

from . import models


@admin.register(models.BankCard)
class BankCardAdmin(admin.ModelAdmin):
    list_display = ('id', 'created', 'user_id', 'is_active', 'bin', 'number', 'expiry_month', 'expiry_year')
    list_filter = ('created',)
    search_fields = ('=id', '=user_id',)
    readonly_fields = ('user_id', 'bin', 'number', 'expiry_month', 'expiry_year', 'token')

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False



from functools import update_wrapper

from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.db import transaction
from django.shortcuts import get_object_or_404
from viewflow.fsm import TransitionNotAllowed


def wrap_task_view(self, origin_view, permission=None):
    """Create wrapper around orig_view to inject request.activation."""

    def view(request, *args, **kwargs):
        process_pk, task_pk = kwargs.get("process_pk"), kwargs.get("task_pk")
        if process_pk is None or task_pk is None:
            raise ImproperlyConfigured(
                "Task view URL path should contain <int:process_pk> and <int:task_pk> parameters"
            )

        def call_with_activation():
            task = get_object_or_404(self.flow_class.task_class, pk=task_pk)
            request.activation = self.activation_class(task)

            if permission and not permission(request.user, task=task):
                raise PermissionDenied

            return origin_view(request, *args, **kwargs)

        try:
            if request.method == "POST":
                with transaction.atomic(), self.flow_class.lock(process_pk):
                    return call_with_activation()
            else:
                return call_with_activation()
        except TransitionNotAllowed as e:
            raise PermissionDenied(",".join(e.args))

    update_wrapper(view, origin_view)
    return view


def wrap_start_view(flow_task, origin_view):
    def view(request, *args, **kwargs):
        request.activation = flow_task.activation_class.create(flow_task, None, None)

        if not flow_task.can_execute(request.user, request.activation.task):
            raise PermissionDenied

        if request.method == "POST":
            with transaction.atomic():
                request.activation.start(request)
                return origin_view(request, *args, **kwargs)
        else:
            request.activation.start(request)
            return origin_view(request, *args, **kwargs)

    update_wrapper(view, origin_view)
    return view


def wrap_view(flow_task, origin_view):
    """Create wrapper around orig_view to inject request.activation."""

    def view(request, *args, **kwargs):
        process_pk, task_pk = kwargs.get("process_pk"), kwargs.get("task_pk")
        if process_pk is None or task_pk is None:
            raise ImproperlyConfigured(
                "Task view URL path should contain <int:process_pk> and <int:task_pk> parameters"
            )

        def call_with_activation():
            task = get_object_or_404(flow_task.flow_class.task_class, pk=task_pk)
            request.activation = flow_task.activation_class(task)

            if not flow_task.can_execute(request.user, task=task):
                raise PermissionDenied

            request.activation.start(request)
            return origin_view(request, *args, **kwargs)

        try:
            if request.method == "POST":
                with flow_task.flow_class.lock(process_pk), transaction.atomic():
                    return call_with_activation()
            else:
                return call_with_activation()
        except TransitionNotAllowed as e:
            raise PermissionDenied(",".join(e.args))

    update_wrapper(view, origin_view)
    return view



""" Standalone workflow views. """

from .actions import (
    AssignTaskView,
    BulkAssignTasksActionView,
    BulkUnassignTasksActionView,
    CancelProcessView,
    CancelTaskView,
    ReviveTaskView,
    UnassignTaskView,
    UndoTaskView,
)
from .chart import FlowChartView
from .create import CreateProcessView, CreateArtifactView
from .dashboard import DashboardView, DashboardTaskListView, DashboardProcessListView
from .detail import IndexTaskView, UserIndexTaskView, DetailTaskView, DetailProcessView
from .mixins import (
    SuccessMessageMixin,
    TaskSuccessUrlMixin,
    ProcessViewTemplateNames,
    TaskViewTemplateNames,
)
from .update import UpdateProcessView, UpdateArtifactView
from .list import (
    FlowInboxListView,
    FlowQueueListView,
    FlowArchiveListView,
    WorkflowInboxListView,
    WorkflowQueueListView,
    WorkflowArchiveListView,
)


__all__ = (
    "AssignTaskView",
    "BulkAssignTasksActionView",
    "BulkUnassignTasksActionView",
    "CancelProcessView",
    "CancelTaskView",
    "CreateArtifactView",
    "CreateProcessView",
    "DashboardProcessListView",
    "DashboardTaskListView",
    "DashboardView",
    "DetailProcessView",
    "DetailTaskView",
    "FlowArchiveListView",
    "FlowChartView",
    "FlowInboxListView",
    "FlowQueueListView",
    "IndexTaskView",
    "ProcessViewTemplateNames",
    "ReviveTaskView",
    "SuccessMessageMixin",
    "TaskSuccessUrlMixin",
    "TaskViewTemplateNames",
    "UnassignTaskView",
    "UndoTaskView",
    "UpdateArtifactView",
    "UpdateProcessView",
    "UserIndexTaskView",
    "WorkflowArchiveListView",
    "WorkflowInboxListView",
    "WorkflowQueueListView",
)



from itertools import count
from django.utils.deconstruct import deconstructible


@deconstructible
class Token(object):
    """
    Helper for tree-like flow token management.

    We follow basic strategy for flow split/join

    Flow starts with only one 'start' token,

    - each split adds split_pk and path id to following task 'start/3_4',
      and 'start/3_4/7_4' on one more split

    - each join removes corresponding split token addition, so 'start/3_4/7_4'
      becomes 'start/3_4'
    """

    def __init__(self, token):
        """
        Instantiate a new token.

        :param token: str
        """
        self.token = token

    def is_split_token(self):
        """True, if it is a token of parallel task."""
        return '/' in self.token

    def get_base_split_token(self):
        """Return token before last split happens."""
        return Token(self.token.rsplit('/', 1)[0])

    def get_common_split_prefix(self, join_token, task_pk):
        """Common prefix for tokens."""
        if self == join_token:
            return '{}/{}_'.format(self.token, task_pk)
        return '{}_'.format(self.token.rsplit('_', 1)[0])

    def __repr__(self):
        return self.token

    def __eq__(self, other):
        if isinstance(other, Token):
            other_token = other.token
        elif isinstance(other, str):
            other_token = other
        else:
            return NotImplemented

        return self.token == other_token

    def __len__(self):
        return len(self.token)

    @classmethod
    def split_token_source(cls, prev_token, task_pk):
        """Span a set of uniq tokens with common prefix."""
        for n in count(1):
            yield Token("{}/{}_{}".format(prev_token, task_pk, n))



from contextlib import contextmanager


from django.db import models
from django.template import Template, Context
from django.utils import timezone
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _

from .fields import FlowReferenceField, TaskReferenceField, TokenField
from .managers import ProcessQuerySet, TaskQuerySet, coerce_to_related_instance
from .token import Token
from . import status


class AbstractProcess(models.Model):
    """Base class for Process data object."""

    flow_class = FlowReferenceField(_("Flow"))
    status = models.CharField(
        _("Status"),
        choices=status.PROCESS.choices,
        default=status.PROCESS.NEW,
        max_length=50,
    )

    created = models.DateTimeField(_("Created"), default=timezone.now)
    finished = models.DateTimeField(_("Finished"), blank=True, null=True)

    objects = ProcessQuerySet.as_manager()

    class Meta:
        abstract = True

    def __str__(self):
        if self.flow_class:
            return "{} #{}".format(self.flow_class.process_title, self.pk)
        return "<Process {}> - {}".format(self.pk, self.status)

    @property
    def brief(self):
        """Quick textual process state representation for end user."""
        template_content = ""

        if self.finished:
            template_content = self.flow_class.process_result_template

        if not template_content:
            template_content = self.flow_class.process_summary_template

        if not template_content:
            template_content = self.flow_class.process_description

        if not template_content:
            template_content = "{{ flow_class.process_title }} - {{ process.status }}"

        return Template(force_str(template_content)).render(
            Context({"process": self.coerced, "flow_class": self.flow_class})
        )

    @property
    def coerced(self):
        """Return process instance of flow_class type."""
        if self.flow_class:
            return coerce_to_related_instance(self, self.flow_class.process_class)
        return self


class AbstractTask(models.Model):
    """
    Base class for Task state objects.

    In addition, you have to define at least process foreign key field::

        process = models.ForeignKey(Process, on_delete=models.CASCADE)

    """

    flow_task = TaskReferenceField(_("Task"))
    flow_task_type = models.CharField(_("Type"), max_length=50)
    status = models.CharField(
        _("Status"),
        choices=status.STATUS.choices,
        db_index=True,
        default=status.STATUS.NEW,
        max_length=50,
    )

    created = models.DateTimeField(_("Created"), default=timezone.now, db_index=True)
    assigned = models.DateTimeField(_("Assigned"), blank=True, null=True)
    started = models.DateTimeField(_("Started"), blank=True, null=True)
    finished = models.DateTimeField(_("Finished"), blank=True, null=True)

    previous = models.ManyToManyField(
        "self", symmetrical=False, related_name="leading", verbose_name=_("Previous")
    )
    token = TokenField(_("Token"), default=Token("start"))

    external_task_id = models.CharField(
        blank=True,
        db_index=True,
        max_length=50,
        null=True,
        verbose_name=_("External Task ID"),
    )

    previous = models.ManyToManyField(
        to="self", symmetrical=False, related_name="leading", verbose_name=_("Previous")
    )

    # permissions
    owner = models.PositiveIntegerField('Сотрудник', null=True, blank=True)

    owner_permission = models.CharField(
        blank=True,
        max_length=255,
        null=True,
        verbose_name=_("Permission"),
    )

    owner_permission_obj_pk = models.CharField(max_length=255, blank=True, null=True)

    objects = TaskQuerySet.as_manager()

    class Meta:
        abstract = True

    def __str__(self):
        if self.flow_task:
            flow_label = self.flow_task.flow_class.instance.flow_label
            return f"<{flow_label}.{self.flow_task}/{self.pk}> - {self.status}"
        return f"<Task {self.pk}> - {self.status}"

    def save(self, *args, **kwargs):  # noqa D102
        if self.flow_task and not self.flow_task_type:
            self.flow_task_type = self.flow_task.task_type

        super(AbstractTask, self).save(*args, **kwargs)

    @property
    def coerced(self):
        """Return task instance of flow_class type."""
        if self.flow_task is not None:
            return coerce_to_related_instance(
                self, self.flow_task.flow_class.task_class
            )

    @property
    def coerced_process(self):
        """Return process instance of flow_class type."""
        if self.flow_task is not None:
            return coerce_to_related_instance(
                self.process, self.flow_task.flow_class.process_class
            )

    @property
    def title(self):
        return self.flow_task.task_title if self.flow_task and self.flow_task.task_title else 'Без названия'

    def brief(self):
        """Quick textual task representation for the end user."""
        if not self.flow_task:
            return "< No flow_task assigned >"

        template_content = ""

        if self.finished:
            template_content = self.flow_task.task_result_template

        if not template_content:
            template_content = self.flow_task.task_summary_template

        if not template_content:
            template_content = self.flow_task.task_description

        if not template_content:
            template_content = self.flow_task.task_title

        if not template_content:
            template_content = "{{ flow_task }}/{{ task.status }}"

        return Template(force_str(template_content)).render(
            Context(
                {
                    "process": self.process.coerced,
                    "task": self.coerced,
                    "flow_class": self.flow_task.flow_class,
                    "flow_task": self.flow_task,
                }
            )
        )

    @contextmanager
    def activation(self):
        """
        Context manager for working with task

        with task.activation() as activation:
            pass
        """
        assert self.pk

        with self.flow_task.flow_class.lock(self.process_id):
            self.refresh_from_db()
            yield self.flow_task.activation_class(self)


class Process(AbstractProcess):
    """Default viewflow Process model."""

    data = models.JSONField(default=dict, blank=True)

    parent_task = models.ForeignKey(
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="+",
        to="Task",
    )

    artifact_object_id = models.PositiveIntegerField(null=True, blank=True)

    class Meta:  # noqa D101
        ordering = ["-created"]
        verbose_name = _("Process")
        verbose_name_plural = _("Process list")


class Task(AbstractTask):
    """Default viewflow Task model."""

    process = models.ForeignKey(
        on_delete=models.CASCADE,
        to=Process,
        verbose_name=_("Process"),
    )

    data = models.JSONField(default=dict, blank=True)

    artifact_object_id = models.PositiveIntegerField(null=True, blank=True)

    class Meta:  # noqa D101
        verbose_name = _("Task")
        verbose_name_plural = _("Tasks")
        ordering = ["-created"]



# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is licensed under the Commercial licence defined in file
# 'COMM_LICENSE', which is part of this source code package.

from django.db.models.query import QuerySet
from rest_framework.views import APIView
from rest_framework.response import Response


class FlowGraphView(APIView):
    flow_state_field = None
    lookup_field = 'pk'
    lookup_url_kwarg = None
    queryset = None

    def get_queryset(self):
        if self.queryset:
            queryset = self.queryset
            if isinstance(queryset, QuerySet):
                # Ensure queryset is re-evaluated on each request.
                queryset = queryset.all()
            return queryset

    def get_object(self):
        queryset = self.get_queryset()
        if queryset is None:
            return

        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field
        if lookup_url_kwarg not in self.kwargs:
            return

        obj = queryset.filter(
            **{self.lookup_field: self.kwargs[lookup_url_kwarg]}
        ).first()
        if not obj:
            return

        self.check_object_permissions(self.request, obj)
        return obj

    def get_flow_graph(self, obj=None):
        raise NotImplementedError

    def get(self, request, format=None):
        obj = self.get_object()
        flow = self.get_flow(obj)
        completed_states = self.get_completed_states(flow, obj)
        graph = self.get_get_flow_graph(completed_states)
        return Response(graph.to_json())



from django.db import models


class BankCard(models.Model):
    created = models.DateTimeField('Создана', auto_now_add=True)
    updated = models.DateTimeField('Изменена', auto_now=True)
    user_id = models.PositiveIntegerField()
    is_active = models.BooleanField('Активна', default=True)
    bin = models.CharField('Бин', max_length=6)
    number = models.CharField('Номер', max_length=4)
    expiry_month = models.PositiveSmallIntegerField('Месяц')
    expiry_year = models.PositiveSmallIntegerField('Год')
    token = models.CharField('Токен', max_length=1024, unique=True)

    def __str__(self):
        return f'{self.bin}****{self.number}'

    class Meta:
        managed = False
        verbose_name_plural = 'Банковские карты'
        verbose_name = 'Банковская карта'
        unique_together = ('bin', 'number', 'expiry_month', 'expiry_year')
        constraints = (
            models.UniqueConstraint(fields=['bin', 'number', 'expiry_month', 'expiry_year'], name='unique_bank_card'),
        )



from django.contrib import messages
from django.views import generic
from django.utils.translation import gettext_lazy as _
from . import mixins


class IndexTaskView(generic.RedirectView):
    """Redirect to the task detail."""

    def get_redirect_url(self, *args, **kwargs):
        activation = self.request.activation

        return activation.flow_task.reverse(
            "detail", args=[activation.process.pk, activation.task.pk]
        )


class UserIndexTaskView(generic.RedirectView):
    """Redirect for a flow.View node."""

    def get_redirect_url(self, *args, **kwargs):
        activation = self.request.activation
        task = activation.task
        flow_task = activation.flow_task

        if activation.start.can_proceed() and flow_task.can_execute(
            self.request.user, task
        ):
            return flow_task.reverse("execute", args=[task.process_id, task.pk])

        if activation.assign.can_proceed() and flow_task.can_assign(
            self.request.user, task
        ):
            return flow_task.reverse("assign", args=[task.process_id, task.pk])

        if flow_task.can_view(self.request.user, task):
            return flow_task.reverse("detail", args=[task.process_id, task.pk])

        messages.success(
            self.request, _("You have no rights to view this task"), fail_silently=True
        )

        return "/"


class DetailTaskView(mixins.TaskViewTemplateNames, generic.TemplateView):
    """
    Default detail view for the flow task.

    Get confirmation from user, assigns task and redirects to task pages.
    """

    template_filename = "task_detail.html"

    def get_actions(self):
        activation = self.request.activation
        return activation.flow_task.get_available_actions(activation, self.request.user)


class DetailProcessView(generic.DetailView):  # todo permission
    """Detail for process."""

    flow_class = None
    context_object_name = "process"
    pk_url_kwarg = "process_pk"

    def get_template_names(self):
        """List of template names to be used for a process detail page.

        If `template_name` is None, default value is::

            [<app_label>/<flow_label>/process_detail.html,
             'viewflow/workflow/process_detail.html']
        """
        if self.template_name is None:
            opts = self.flow_class.instance

            return (
                "{}/{}/process_detail.html".format(opts.app_label, opts.flow_label),
                "viewflow/workflow/process_detail.html",
            )
        else:
            return [self.template_name]

    def get_context_data(self, **kwargs):
        """Context for a detail view.

        :keyword process: a Process instance
        :keyword task_list: List of tasks of the process
        """
        context = super(DetailProcessView, self).get_context_data(**kwargs)
        context["task_list"] = context["process"].task_set.all().order_by("created")
        return context

    def get_queryset(self):
        """Return the `QuerySet` that will be used to look up the process."""
        if self.queryset is None:
            return self.flow_class.process_class._default_manager.all()
        return self.queryset.all()


class DetailSubprocessView(DetailTaskView):
    def get_template_names(self):
        flow_task = self.activation.flow_task
        opts = self.activation.flow_task.flow_class.instance

        return (
            "{}/{}/{}_detail.html".format(
                opts.app_label, opts.flow_label, flow_task.name
            ),
            "{}/{}/subprocess_detail.html".format(opts.app_label, opts.flow_label),
            "viewflow/workflow/subprocess_detail.html",
        )



# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.

from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from viewflow import Icon
from viewflow.urls import Application


class Admin(Application):
    """
    Django administration Viewset adapter::

        from material.contrib.admin import Admin

        site = Site(apps=[
            Admin()
        ])

        urls = [
            path('', site.urls)
        ]

    """
    app_name = 'admin'
    namespace = 'admin'
    prefix = 'admin'

    title = _("Administration")
    icon = Icon("build")
    turbo_disabled = True

    def __init__(self, *, admin_site=None, **kwargs):
        self.admin_site = admin_site or admin.site
        super().__init__(**kwargs)

    def has_perm(self, user):
        return user.is_staff

    @property
    def urls(self):
        url_patterns, app_name, namespace = self.admin_site.urls
        assert self.app_name == app_name
        assert self.namespace == namespace
        return url_patterns, app_name, namespace



from __future__ import unicode_literals

from django.core.management.base import BaseCommand, CommandError
from viewflow.workflow import chart
from viewflow.workflow.fields import import_flow_by_ref


class Command(BaseCommand):
    help = "Export flow graph"

    def add_arguments(self, parser):
        parser.add_argument(
            "flow_label",
            nargs=1,
            type=str,
            help="Flow label, i.e. app_label/flows.MyFlow",
        )

        parser.add_argument(
            "--output",
            "-o",
            action="store",
            dest="output",
            help="Render output to a file",
        )

        parser.add_argument(
            "--process_pk",
            "-pk",
            action="store",
            dest="process",
            help="Draw process instance state",
        )

        parser.add_argument(
            "--format",
            dest="format",
            action="store",
            default="svg",
            help="Export format (svg or bpmn)",
        )

    def handle(self, **options):
        flow_class = import_flow_by_ref(options["flow_label"][0])

        result = ""
        process_pk = options.get("process")
        output_format = options.get("format")
        if output_format == "svg":
            grid = chart.calc_layout_data(flow_class)
            if process_pk is not None:
                chart.calc_cell_status(flow_class, grid, process_pk)
            result = chart.grid_to_svg(grid)
        elif output_format == "bpmn":
            grid = chart.calc_layout_data(
                flow_class, edge_start_gap_size=0, edge_end_gap_size=0
            )
            if process_pk is not None:
                chart.calc_cell_status(flow_class, grid, process_pk)
            result = chart.grid_to_bpmn(grid)
        else:
            CommandError("Unknown output format {}".format(output_format))

        filename = options.get("output")
        if filename:
            with open(filename, "w") as output:
                output.write(result)
        else:
            print(result)



from django.contrib import admin

from . import models


@admin.register(models.Profile)
class ProfileAdmin(admin.ModelAdmin):
    list_display = ('created', 'user_id', 'last_name', 'first_name', 'middle_name', 'birth_date',
                    'identified', 'verified',
                    'passport_serial', 'passport_number', 'snils'
                    )
    search_fields = ('=user_id', '=passport_number', '=snils', '=last_name', '=first_name', '=middle_name')
    list_filter = ('created', 'identified', 'verified')
    ordering = ('-created',)
    readonly_fields = ('user_id',)

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False



"""Base workflow node set."""

from .end import End, EndActivation
from .func import Function, FunctionActivation
from .handle import Handle, HandleActivation
from .if_gate import If, IfActivation
from .job import AbstractJob, AbstractJobActivation
from .join import Join, JoinActivation
from .obsolete import Obsolete, ObsoleteActivation
from .split import Split, SplitActivation, SplitFirst
from .start import Start, StartHandle, StartActivation
from .subprocess import StartSubprocess, Subprocess, NSubprocess
from .switch import Switch, SwitchActivation
from .view import ViewActivation, View

__all__ = (
    "AbstractJob",
    "AbstractJobActivation",
    "End",
    "EndActivation",
    "Function",
    "FunctionActivation",
    "Handle",
    "HandleActivation",
    "If",
    "IfActivation",
    "Join",
    "JoinActivation",
    "NSubprocess",
    "Obsolete",
    "ObsoleteActivation",
    "Split",
    "SplitActivation",
    "SplitFirst",
    "Start",
    "StartActivation",
    "StartHandle",
    "StartSubprocess",
    "Subprocess",
    "Switch",
    "SwitchActivation",
    "View",
    "ViewActivation",
)



from django.contrib import admin

from .models import BlackList


@admin.register(BlackList)
class BlackListAdmin(admin.ModelAdmin):
    list_display = ('created', 'user_id', 'passport_serial', 'passport_number', 'cause', 'comment')
    search_fields = ('=user_id', '=passport_number')
    list_filter = ('cause',)
    ordering = ('-created', )



# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.

import copy
import json
from datetime import date, datetime
from functools import partialmethod

from django.core.exceptions import FieldError
from django.db.models import fields, DEFERRED
from django.db.models import JSONField as BaseJSONField
from django.utils import dateparse, timezone


DEFAULT = object()


class JSONFieldDescriptor(object):
    def __init__(self, field):
        self.field = field

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        json_value = getattr(instance, self.field.json_field_name)
        if isinstance(json_value, dict):
            if self.field.attname in json_value:
                value = json_value.get(self.field.attname, None)
                if hasattr(self.field, "from_json"):
                    value = self.field.from_json(value)
            elif self.field.default and self.field.default != fields.NOT_PROVIDED:
                value = self.field.default
            else:
                value = None
            return value
        return None

    def __set__(self, instance, value):
        json_value = getattr(instance, self.field.json_field_name)
        if json_value:
            assert isinstance(json_value, dict)
        else:
            json_value = {}

        if hasattr(self.field, "to_json"):
            value = self.field.to_json(value)

        if not value and self.field.blank and not self.field.null:
            try:
                del json_value[self.field.attname]
            except KeyError:
                pass
        else:
            json_value[self.field.attname] = value

        setattr(instance, self.field.json_field_name, json_value)


class JSONFieldMixin(object):
    """
        Override django.db.model.fields.Field.contribute_to_class
    to make a field always private, and register custom access descriptor
    """

    def __init__(self, *args, **kwargs):
        self.json_field_name = kwargs.pop("json_field_name", "data")
        super(JSONFieldMixin, self).__init__(*args, **kwargs)

    def contribute_to_class(self, cls, name, private_only=False):
        self.set_attributes_from_name(name)
        self.model = cls
        self.concrete = False
        self.column = None  # self.json_field_name
        cls._meta.add_field(self, private=True)

        if not getattr(cls, self.attname, None):
            descriptor = JSONFieldDescriptor(self)
            setattr(cls, self.attname, descriptor)

        if self.choices is not None:
            setattr(
                cls,
                "get_%s_display" % self.name,
                partialmethod(cls._get_FIELD_display, field=self),
            )

        self.column = self.json_field_name

        # def on_model_init(sender, name=name, **kwargs):
        #     """ Initialize default values """
        #     instance = kwargs['instance']
        #     if getattr(instance, name) is None:
        #         field = instance._meta.get_field(name)
        #         if not field.null:
        #             setattr(instance, name, field.get_default())

        # post_init.connect(on_model_init, sender=cls, weak=False)

    # tOod why?
    def get_default(self):
        return DEFERRED

    def get_lookup(self, lookup_name):
        # Always return None, to make get_transform been called
        return None

    def formfield(self, **kwargs):
        if self.has_default():
            if not callable(self.default):
                kwargs["initial"] = self._get_default()
        return super().formfield(**kwargs)

    def get_transform(self, name):
        class TransformFactoryWrapper:
            def __init__(self, json_field, transform, original_lookup):
                self.json_field = json_field
                self.transform = transform
                self.original_lookup = original_lookup

            def __call__(self, lhs, **kwargs):
                lhs = copy.copy(lhs)
                lhs.target = self.json_field
                lhs.output_field = self.json_field
                transform = self.transform(lhs, **kwargs)
                transform._original_get_lookup = transform.get_lookup
                transform.get_lookup = lambda name: transform._original_get_lookup(
                    self.original_lookup
                )
                return transform

        json_field = self.model._meta.get_field(self.json_field_name)
        transform = json_field.get_transform(self.name)
        if transform is None:
            raise FieldError(
                "JSONField '%s' has no support for key '%s' %s lookup"
                % (self.json_field_name, self.name, name)
            )

        return TransformFactoryWrapper(json_field, transform, name)

    def convert_json_value(self, value, expression, connection):
        if isinstance(value, str):
            return json.loads(value)[self.name]
        return value

    def get_db_converters(self, connection):
        converters = super().get_db_converters(connection)
        converters += [self.convert_json_value]
        return converters


class BooleanField(JSONFieldMixin, fields.BooleanField):
    def __init__(self, *args, **kwargs):
        super(BooleanField, self).__init__(*args, **kwargs)
        self.blank = False


class CharField(JSONFieldMixin, fields.CharField):
    pass


class DateField(JSONFieldMixin, fields.DateField):
    def to_json(self, value):
        if value:
            assert isinstance(value, (datetime, date))
            return value.strftime("%Y-%m-%d")

    def from_json(self, value):
        if value is not None:
            return dateparse.parse_date(value)


class DateTimeField(JSONFieldMixin, fields.DateTimeField):
    def to_json(self, value):
        if value:
            if not timezone.is_aware(value):
                value = timezone.make_aware(value)
            return value.isoformat()

    def from_json(self, value):
        if value:
            return dateparse.parse_datetime(value)


class DecimalField(JSONFieldMixin, fields.DecimalField):
    pass


class EmailField(JSONFieldMixin, fields.EmailField):
    pass


class FloatField(JSONFieldMixin, fields.FloatField):
    pass


class IntegerField(JSONFieldMixin, fields.IntegerField):
    pass


class IPAddressField(JSONFieldMixin, fields.IPAddressField):
    pass


class GenericIPAddressField(JSONFieldMixin, fields.GenericIPAddressField):
    pass


class NullBooleanField(JSONFieldMixin, fields.NullBooleanField):
    pass


class TextField(JSONFieldMixin, fields.TextField):
    pass


class TimeField(JSONFieldMixin, fields.TimeField):
    def to_json(self, value):
        if value:
            if not timezone.is_aware(value):
                value = timezone.make_aware(value)
            return value.isoformat()

    def from_json(self, value):
        if value:
            return dateparse.parse_time(value)


class URLField(JSONFieldMixin, fields.URLField):
    pass


class JSONField(JSONFieldMixin, BaseJSONField):
    pass



# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.

import re
from typing import Any, Callable, Iterator, List, Optional, Tuple, Type, TypeVar

from django.apps import apps
from django.conf import settings
from django.contrib import auth
from django.contrib.auth import get_permission_codename
from django.core import mail
from django.db import models, ProgrammingError
from django.db.models import ForeignObjectRel, QuerySet
from django.urls import reverse, NoReverseMatch
from django.utils.html import conditional_escape
from django.utils.safestring import mark_safe

__all__ = ("has_object_perm", "viewprop", "DEFAULT", "first_not_default")


T = TypeVar("T")
TCallable = TypeVar("TCallable", bound=Callable[..., Any])


class MARKER(object):
    def __init__(self, marker: str):
        self.marker = marker

    def __iter__(self):
        return iter(())

    def __repr__(self):
        return self.marker


DEFAULT = MARKER("DEFAULT")

IS_DEV = settings.DEBUG or not hasattr(mail, "outbox")  # DEBUG or test mode


def first_not_default(*args):
    """
    Return the first argument that is not the `DEFAULT` marker. If all arguments
    are `DEFAULT`, return the last one.
    """
    if not args:
        return None

    for arg in args:
        if arg is not DEFAULT:
            return arg
    return arg


def camel_case_to_underscore(name):
    """
    Convert a camel-cased string to an underscore-separated string.
    For example, 'SomeString' becomes 'some_string'.
    """

    return re.sub(
        "([a-z0-9])([A-Z])", r"\1_\2", re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    ).lower()


def camel_case_to_title(name):
    """
    Convert a camel-cased string to a title-cased string.
    For example, 'SomeString' becomes 'Some String'.
    """

    return re.sub(
        "([a-z0-9])([A-Z])", r"\1 \2", re.sub("(.)([A-Z][a-z]+)", r"\1 \2", name)
    ).capitalize()


def has_object_perm(user, short_perm_name, model, obj=None):
    """
    Check if the user has the specified permission for the given model. If an
    object is provided, and user has no model-wide permission, check if the user
    has the permission for that specific object instance.
    """
    perm_name = f"{model._meta.app_label}.{auth.get_permission_codename(short_perm_name, model._meta)}"
    has_perm = user.has_perm(perm_name)
    if not has_perm and obj is not None:
        has_perm = user.has_perm(perm_name, obj=obj)
    return has_perm


def strip_suffixes(word, suffixes):
    """
    Strip the specified suffixes from the given word.
    Never strip the whole word to an empty string.
    """

    for suffix in suffixes:
        if word != suffix and word.endswith(suffix):
            word = word[: -len(suffix)]
    return word


def get_app_package(app_label):
    """
    Returns the name of the package that contains the specified app or None if
    the app is not found.
    """
    app_config = apps.get_app_config(app_label)
    if not app_config:
        return None
    return app_config.module.__name__


def get_containing_app_data(module):
    """
    Returns the app label and package string for the specified module.
    """
    app_config = apps.get_containing_app_config(module)
    if not app_config:
        return None, None
    return app_config.label, app_config.module.__name__


def is_owner(owner: models.Model, user: models.Model) -> bool:
    """
    Checks whether the specified user instance or subclass is equal to the
    specified owner instance or subclass.
    """
    return owner == user.pk


class viewprop:
    """
    A property that can be overridden.

    The viewprop class is a descriptor that works similarly to the built-in
    `property` decorator but allows its value to be overridden on instances
    of the class it is used in.
    """

    def __init__(self, func: Any):
        self.__doc__ = getattr(func, "__doc__")
        self.fget = func

    def __get__(self, obj: Optional[Any], objtype: Optional[Type[Any]] = None) -> Any:
        if obj is None:
            return self
        if self.fget.__name__ not in obj.__dict__:
            obj.__dict__[self.fget.__name__] = self.fget(obj)
        return obj.__dict__[self.fget.__name__]

    def __set__(self, obj: Any, value: Any) -> None:
        obj.__dict__[self.fget.__name__] = value

    def __repr__(self) -> str:
        return "<view_property func={}>".format(self.fget)


class LazySingletonDescriptor:
    """
    Descriptor class that creates a lazy singleton instance.

    This descriptor can be used as a class attribute, and the first time the
    attribute is accessed, it creates an instance of the class. Subsequent
    accesses return the same instance, effectively making the class a singleton.
    """

    def __init__(self) -> None:  # noqa D102
        self.instance: Optional[T] = None

    def __get__(
        self,
        instance: Optional[T] = None,
        owner: Optional[Type[T]] = None,
    ) -> T:
        if self.instance is None:
            if owner is None:
                raise ValueError("Owner class not provided")
            self.instance = owner()
        return self.instance


class Icon:
    """
    Class representing an HTML icon element.

    Attributes:
    -----------
    icon_name : str
        The name of the icon to use.
    class_ : str, optional
        The CSS class to apply to the icon element.
    """

    def __init__(self, icon_name: str, class_: Optional[str] = None):
        self.icon_name = icon_name
        self.class_ = class_ or ""

    def __str__(self) -> str:
        icon_name = conditional_escape(self.icon_name)
        class_name = conditional_escape(self.class_)
        return mark_safe(
            f'<i class="material-icons ${class_name}" aria-hidden="true">{icon_name}</i>'
        )


def get_object_data(obj: models.Model) -> Iterator[Tuple[models.Field, str, Any]]:
    """
    List of object fields to display. Choice fields values are expanded to
    readable choice label.

    Returns a list of (field, label, value) tuples for the fields of the given
    object.

    """
    for field in obj._meta.fields:
        if isinstance(field, models.AutoField):
            continue
        elif field.auto_created:
            continue
        else:
            choice_display_attr = "get_{}_display".format(field.name)
        if hasattr(obj, choice_display_attr):
            value = getattr(obj, choice_display_attr)()
        else:
            value = getattr(obj, field.name)

        if value is not None:
            yield (field, field.verbose_name.capitalize(), value)

    if hasattr(obj, "artifact_object_id") and obj.artifact_object_id:
        yield from get_object_data(obj.artifact)


PATH_PARAMETER_COMPONENT_RE = re.compile(
    r"<(?:(?P<converter>[^>:]+):)?(?P<parameter>[^>]+)>"
)


def list_path_components(route: str) -> List[str]:
    """
    Extract keyword arguments from a Django path expression, which are used as
    input parameters for a view function.

    Example Usage:

     >>> list_path_components('/prefix/<str:pk>')
     ['pk']
     >>> list_path_components('<str:pk>/<int:id>')
     ['pk', 'id']
    """
    return [match["parameter"] for match in PATH_PARAMETER_COMPONENT_RE.finditer(route)]


def get_model_display_data(root_instance, user):
    """
    Return structure with model fields and related from same app.

    Example:

        [(Title, [(Field Title, Value), ... ]), ...]

    """
    result = []
    new_objects = [(root_instance._meta.verbose_name.title(), root_instance)]
    processed_models, processed_objects = [], []

    def expand_required(instance):
        if instance in processed_objects:
            return False
        if instance.__class__ in processed_models:
            return False
        return (root_instance._meta.app_label == instance._meta.app_label)

    # Ext models
    models_names = root_instance.add_models if hasattr(root_instance, 'add_models') else tuple()
    for name in models_names:
        if hasattr(root_instance, name):
            objs = getattr(root_instance, name)
            if objs and isinstance(objs, models.Model):
                new_objects.append((objs._meta.verbose_name.title(), objs))
            elif objs and isinstance(objs, QuerySet):
                for obj in objs:
                    new_objects.append((obj._meta.verbose_name.title(), obj))

    while new_objects:
        root_title, root = new_objects.pop(0)
        root_admin_url = None
        children = []

        processed_objects.append(root)
        processed_models.append(root.__class__)

        # objects fields
        for field in root._meta.fields:
            # if isinstance(field, models.AutoField):
            #     continue
            if field.auto_created:
                continue
            elif field.name == 'data':
                continue
            elif isinstance(field, models.ForeignKey) and not field.auto_created:
                related_id = getattr(root, field.get_attname())
                if related_id is not None:
                    related = getattr(root, field.name)
                    if expand_required(related):
                        if root_instance._meta.app_label == related._meta.app_label:
                            new_objects.append((field.verbose_name.title(), related))
                        else:
                            value = getattr(root, field.name)
                            if value is not None:
                                children.append((field.verbose_name.title(), value))
            else:
                choice_display_attr = "get_{}_display".format(field.get_attname())
                if hasattr(root, choice_display_attr):
                    value = getattr(root, choice_display_attr)()
                else:
                    value = getattr(root, field.get_attname())

                if value is not None:
                    children.append((field.verbose_name.title(), value))

        # backward relations
        backward_relations = [
            field for field in root._meta.get_fields()
            if isinstance(field, ForeignObjectRel)
        ]
        try:
            for relation in backward_relations:
                if not isinstance(relation.field, models.OneToOneField):
                    for related in getattr(root, relation.get_accessor_name()).all():
                        if expand_required(related):
                            new_objects.append((related._meta.verbose_name.title(), related))
        except ProgrammingError:
            pass

        for name in models_names:
            if hasattr(root_instance, name):
                objs = getattr(root_instance, name)
                if objs.__class__ == root.__class__:
                    for property_name in models_names[name]:
                        if hasattr(root, property_name):
                            children.append((property_name, getattr(root, property_name)))

        # if any suitable for display children found
        if children:
            change_perm = get_permission_codename('change', root._meta)
            if user.has_perm("%s.%s" % (root._meta.app_label, change_perm)):
                admin_url_name = "admin:{}_{}_change".format(root._meta.app_label, root._meta.model_name)
                try:
                    root_admin_url = reverse(admin_url_name, args=(root.pk,))
                except NoReverseMatch:
                    pass
            if root_instance._meta.verbose_name.title() != root_title:
                result.append((root.pk, root_title, children, root_admin_url))

    return result



# coding: utf-8
from __future__ import division

import math
from copy import copy
from itertools import groupby
from textwrap import wrap

from django.template.loader import get_template
from django.utils.encoding import force_str
from .status import STATUS

GAP_SIZE = 25
EDGE_START_GAP_SIZE = 2
EDGE_END_GAP_SIZE = 7

DEFAULT_SHAPE = {
    'width': 150,
    'height': 100,
    'svg': """
        <rect class="task" width="150" height="100" rx="5" ry="5"/>
    """
}


class Edge(object):
    __slots__ = ['src', 'dst', 'segments']

    def __init__(self, src, dst, segments=None):
        self.src = src
        self.dst = dst
        self.segments = segments if segments is not None else []


class Shape(object):
    __slots__ = ['x', 'y', 'width', 'height', 'svg', 'text']

    def __init__(self, x=-1, y=-1, width=-1, height=-1):
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.text = []

    def label(self):
        return ' '.join(segment[0] for segment in self.text)


class Cell(object):
    __slots__ = ['col', 'row', 'x', 'y', 'width', 'height', 'node', 'shape', 'status']

    def __init__(self, node, col=-1, row=-1, x=-1, y=-1, width=-1, height=-1, shape=None, status=None):
        self.node = node
        self.col = col
        self.row = row
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.shape = shape if shape is not None else Shape()
        self.status = status

    def incoming(self):
        return self.node._incoming()

    def outgoing(self):
        return self.node._outgoing()


class Grid(object):
    def __init__(self, nodes):
        self.width = -1
        self.height = -1

        self.grid = {
            node: Cell(node)
            for node in nodes
        }
        self.edges = [
            Edge(edge.src, edge.dst)
            for node in nodes
            for edge in node._incoming()]

    def __getitem__(self, node):
        return self.grid[node]

    @property
    def nodes(self):
        return self.grid.keys()

    @property
    def cells(self):
        return self.grid.values()

    def insert_row_below(self, row):
        for cell in self.cells:
            if cell.row != -1 and cell.row > row:
                cell.row += 1

    def insert_row_above(self, row):
        for cell in self.cells:
            if cell.row != -1 and cell.row >= row:
                cell.row += 1

    def place_start_node(self, node):
        """Place node on a new row in a first column."""
        cell = self[node]
        last_row = max(cell.row for cell in self.cells)
        cell.col, cell.row = 0, last_row + 1

    def place_next_to(self, prev, node):
        """Place node at the right of the previous node.

        If node have no row pre-initialized, the row of previous node is used
        """
        prev_cell, node_cell = self[prev], self[node]
        if node_cell.row == -1:
            node_cell.row = prev_cell.row
        node_cell.col = prev_cell.col + 1

    def distribute_children(self, parent, children):
        """Preinitialize children node rows to be equally spread over parent."""
        children = [child for child in children if self[child].row == -1]
        parent_cell = self[parent]
        half = len(children) / 2
        above = children[0: int(math.floor(half))]
        middle = children[int(math.floor(half)):int(math.ceil(half))]
        below = children[int(math.ceil(half)):]
        for child in above:
            self.insert_row_above(parent_cell.row)
            self[child].row = parent_cell.row - 1
        for child in middle:
            self[child].row = parent_cell.row
        for child in reversed(below):
            self.insert_row_below(parent_cell.row)
            self[child].row = parent_cell.row + 1

    def place_join(self, prev_split, prevs, node):
        """Place join after prev nodes on the same row as prev_split.

        If prev_split is None, center row of prev nodes used
        """
        node_cell = self[node]
        node_cell.col = max(self[prev].col for prev in prevs) + 1
        if prev_split:
            prev_split_cell = self[prev_split]
            node_cell.row = prev_split_cell.row
        else:
            if node_cell.row == -1:
                max_prev_row = max(self[prev].row for prev in prevs)
                min_prev_row = min(self[prev].row for prev in prevs)
                row = (max_prev_row - min_prev_row) / 2
                if row % 1 == 0:
                    node_cell.row = int(row)
                else:
                    self.insert_row_below(int(row))
                    node_cell.row = int(row) + 1

    def collapse_grid(self):
        """Optimize grid by removing empty cells in between."""
        # TODO


def topsort(flow_class):
    result, incoming_to_reverse = [], {}

    nodes = list(flow_class.instance.nodes())
    incoming_edges = {
        node: {edge.src for edge in node._incoming()}
        for node in nodes
    }
    initial_incoming = {
        node: set(incoming_edges[node])
        for node in nodes
    }

    while nodes:
        start_nodes, circle_broken = [], False
        for node in nodes:
            if not incoming_edges[node]:
                start_nodes.append(node)

        if start_nodes:
            for start_node in start_nodes:
                nodes.remove(start_node)
                result.append(start_node)
                for _, sources in incoming_edges.items():
                    sources.discard(start_node)
        else:
            # Fix edges
            for join in nodes:
                if len(incoming_edges[join]) < len(initial_incoming[join]):
                    incoming_to_reverse[join] = copy(incoming_edges[join])
                    for source in incoming_edges[join]:
                        incoming_edges[source].add(join)
                    incoming_edges[join] = set()
                    circle_broken = True
            if not circle_broken:
                node = nodes[0]  # TODO Grab node from a circle
                nodes.remove(node)
                result.append(node)
                for _, sources in incoming_edges.items():
                    sources.discard(node)

    return result, incoming_to_reverse


def find_prev_split(nodes, incoming_edges, outgoing_edges, join):
    """Find the split for join"""
    def is_split_node(source):
        return len(outgoing_edges[source]) > 1

    def is_join_node(source):
        return len(incoming_edges[source]) > 1

    def collect_splits(node, prev_nodes=None):
        """Breadth first split node list."""
        if prev_nodes is None:
            prev_nodes = []

        for source in incoming_edges[node]:
            if is_split_node(source):
                yield source
        for source in incoming_edges[node]:
            if source not in prev_nodes:
                for split in collect_splits(source, prev_nodes + [source]):
                    yield split

    incoming_splits = [
        [source] if is_split_node(source) else [] + list(collect_splits(source))
        for source in incoming_edges[join]
    ]
    if not incoming_splits:
        # No incoming edges
        return None

    common_splits = sorted(
        set.intersection(*[set(splits) for splits in incoming_splits]),
        key=lambda split: incoming_splits[0].index(split)
    )

    if not common_splits:
        return None

    return common_splits[0]


def layout(nodes, incoming_edges, outgoing_edges):
    grid = Grid(nodes)

    for node in nodes:
        incomings, outgoings = incoming_edges[node], outgoing_edges[node]
        placed_incomings = [incoming for incoming in incomings if grid[incoming].col != -1]

        if len(placed_incomings) == 0:
            grid.place_start_node(node)
        elif len(placed_incomings) == 1:
            grid.place_next_to(incomings[0], node)
        else:
            prev_split = find_prev_split(nodes, incoming_edges, outgoing_edges, node)
            grid.place_join(prev_split, placed_incomings, node)

        if len(outgoings) > 1:
            for child in outgoings:
                if len(incoming_edges[child]) > 1:
                    if node in incoming_edges[child]:
                        # join connected directly to split
                        outgoings = copy(outgoings)
                        outgoings.remove(child)
                        outgoings = outgoings[0:int(len(outgoings) / 2)] + [child] + outgoings[int(len(outgoings) / 2):]
                        break
            grid.distribute_children(node, outgoings)

    grid.collapse_grid()

    return grid


def calc_layout_data(flow_class, edge_start_gap_size=EDGE_START_GAP_SIZE, edge_end_gap_size=EDGE_END_GAP_SIZE):
    # Topological nodes sort, and edge cycle breaking
    nodes, _ = topsort(flow_class)

    # Topologically sorted incoming edges
    incoming_edges = {
        node: sorted(
            [edge.src for edge in node._incoming()],
            key=lambda node: nodes.index(node))
        for node in nodes
    }

    # Topologically sorted outgoing edges
    outgoing_edges = {
        node: sorted(
            [edge.dst for edge in node._outgoing()],
            key=lambda node: nodes.index(node))
        for node in nodes
    }

    # Place nodes on a grid
    grid = layout(nodes, incoming_edges, outgoing_edges)

    # hack fix
    for cell1 in grid.cells:
        for cell2 in grid.cells:
            if cell1.node != cell2.node and cell1.row == cell2.row and cell1.col == cell2.col:
                cell1.row += 1

    init_shapes(grid)
    calc_edges(grid, edge_start_gap_size=edge_start_gap_size, edge_end_gap_size=edge_end_gap_size)
    calc_text(grid)

    return grid


def init_shapes(grid):
    # init shapes
    for cell in grid.cells:
        shape = getattr(cell.node, 'shape', DEFAULT_SHAPE)
        cell.shape.width = shape['width']
        cell.shape.height = shape['height']
        cell.shape.svg = shape['svg']

    # calc rows and col sizes
    row_sizes = [0] * max(cell.row + 1 for cell in grid.cells)
    col_sizes = [0] * max(cell.col + 1 for cell in grid.cells)
    for cell in grid.cells:
        row_sizes[cell.row] = max(row_sizes[cell.row], cell.shape.height)
        col_sizes[cell.col] = max(col_sizes[cell.col], cell.shape.width)

    # set shapes positions
    for cell in grid.cells:
        cell.x = sum(col_sizes[:cell.col]) + GAP_SIZE * cell.col
        cell.y = sum(row_sizes[:cell.row]) + GAP_SIZE * cell.row

        cell.width = col_sizes[cell.col]
        if cell.shape.width < col_sizes[cell.col]:
            cell.shape.x = cell.x + int((col_sizes[cell.col] - cell.shape.width) / 2)
        else:
            # is it ever happens?
            cell.shape.x = cell.x

        cell.height = row_sizes[cell.row]
        if cell.shape.height < row_sizes[cell.row]:
            cell.shape.y = cell.y + int((row_sizes[cell.row] - cell.shape.height) / 2)
        else:
            cell.shape.y = cell.y

    grid.width = sum(col_sizes) + GAP_SIZE * len(col_sizes)
    grid.height = sum(row_sizes) + GAP_SIZE * len(row_sizes)
    return grid


def calc_edges(grid, edge_start_gap_size=EDGE_START_GAP_SIZE, edge_end_gap_size=EDGE_END_GAP_SIZE):
    for edge in grid.edges:
        src_cell = grid[edge.src]
        dst_cell = grid[edge.dst]

        if src_cell.row == dst_cell.row and src_cell.col < dst_cell.col:  # →
            edge.segments = [
                [
                    src_cell.shape.x + src_cell.shape.width + edge_start_gap_size,
                    src_cell.shape.y + int(src_cell.shape.height / 2)
                ],
                [
                    dst_cell.shape.x - edge_end_gap_size,
                    dst_cell.shape.y + int(dst_cell.shape.height / 2)
                ]
            ]
        elif src_cell.row < dst_cell.row and src_cell.col < dst_cell.col:  # ↘
            if len(list(edge.src._outgoing())) > 1:
                edge.segments = [
                    [
                        src_cell.shape.x + int(src_cell.shape.width / 2),
                        src_cell.shape.y + src_cell.shape.height + edge_start_gap_size
                    ],
                    [
                        src_cell.shape.x + int(src_cell.shape.width / 2),
                        dst_cell.shape.y + int(dst_cell.shape.height / 2),
                    ],
                    [
                        dst_cell.shape.x - edge_end_gap_size,
                        dst_cell.shape.y + int(dst_cell.shape.height / 2),
                    ]
                ]
            else:
                edge.segments = [
                    [
                        src_cell.shape.x + src_cell.shape.width + edge_start_gap_size,
                        src_cell.shape.y + int(src_cell.shape.height / 2)
                    ],
                    [
                        dst_cell.shape.x + int(dst_cell.shape.width / 2),
                        src_cell.shape.y + int(src_cell.shape.height / 2)
                    ],
                    [
                        dst_cell.shape.x + int(dst_cell.shape.width / 2),
                        dst_cell.shape.y - edge_end_gap_size
                    ]
                ]
        elif src_cell.row < dst_cell.row and src_cell.col == dst_cell.col:  # ↓
            edge.segments = [
                [
                    src_cell.shape.x + int(src_cell.shape.width / 2),
                    src_cell.shape.y + src_cell.shape.height + edge_start_gap_size
                ],
                [
                    dst_cell.shape.x + int(dst_cell.shape.width / 2),
                    dst_cell.shape.y - edge_end_gap_size
                ]
            ]
        elif src_cell.row < dst_cell.row and src_cell.col > dst_cell.col:  # ↙
            edge.segments = [
                [
                    src_cell.shape.x + int(src_cell.shape.width / 2),
                    src_cell.shape.y + src_cell.shape.height + edge_start_gap_size
                ],
                [
                    src_cell.shape.x + int(src_cell.shape.width / 2),
                    src_cell.y + src_cell.height + GAP_SIZE // 2
                ],
                [
                    dst_cell.shape.x + int(3 * dst_cell.shape.width / 4),
                    src_cell.y + src_cell.height + GAP_SIZE // 2
                ],
                [
                    dst_cell.shape.x + int(3 * dst_cell.shape.width / 4),
                    dst_cell.shape.y - edge_end_gap_size
                ]
            ]
        elif src_cell.row == dst_cell.row and src_cell.col > dst_cell.col:  # ←
            edge.segments = [
                [
                    src_cell.shape.x + int(src_cell.shape.width / 2),
                    src_cell.shape.y + src_cell.shape.height + edge_start_gap_size
                ],
                [
                    src_cell.shape.x + int(src_cell.shape.width / 2),
                    src_cell.y + src_cell.height + int(4 * GAP_SIZE // 5)
                ],
                [
                    dst_cell.shape.x + int(3 * dst_cell.shape.width / 4),
                    src_cell.y + src_cell.height + int(4 * GAP_SIZE // 5)
                ],
                [
                    dst_cell.shape.x + int(3 * dst_cell.shape.width / 4),
                    dst_cell.shape.y + dst_cell.shape.height + edge_end_gap_size
                ]
            ]
        elif src_cell.row > dst_cell.row and src_cell.col > dst_cell.col:  # ↖
            edge.segments = [
                [
                    src_cell.shape.x + int(src_cell.shape.width / 2),
                    src_cell.shape.y - edge_start_gap_size
                ],
                [
                    src_cell.shape.x + int(src_cell.shape.width / 2),
                    src_cell.y - GAP_SIZE // 2
                ],
                [
                    dst_cell.shape.x + int(3 * dst_cell.shape.width / 4),
                    src_cell.y - GAP_SIZE // 2
                ],
                [
                    dst_cell.shape.x + int(3 * dst_cell.shape.width / 4),
                    dst_cell.shape.y + dst_cell.shape.height + edge_end_gap_size
                ]
            ]
        elif src_cell.row > dst_cell.row and src_cell.col == dst_cell.col:  # ↑
            edge.segments = [
                [
                    src_cell.shape.x + int(src_cell.shape.width / 2),
                    src_cell.shape.y - edge_start_gap_size
                ],
                [
                    dst_cell.shape.x + int(dst_cell.shape.width / 2),
                    dst_cell.shape.y + dst_cell.shape.height + edge_end_gap_size
                ]
            ]
        elif src_cell.row > dst_cell.row and src_cell.col < dst_cell.col:  # ↗
            if len(list(edge.src._outgoing())) > 1:
                edge.segments = [
                    [
                        src_cell.shape.x + int(src_cell.shape.width / 2),
                        src_cell.shape.y - edge_start_gap_size
                    ],
                    [
                        src_cell.shape.x + int(src_cell.shape.width / 2),
                        dst_cell.shape.y + int(dst_cell.shape.height / 2)
                    ],
                    [
                        dst_cell.shape.x - edge_end_gap_size,
                        dst_cell.shape.y + int(dst_cell.shape.height / 2)
                    ]
                ]
            else:
                edge.segments = [
                    [
                        src_cell.shape.x + src_cell.shape.width + edge_start_gap_size,
                        src_cell.shape.y + int(src_cell.shape.height / 2)
                    ],
                    [
                        dst_cell.shape.x + int(dst_cell.shape.width / 2),
                        src_cell.shape.y + int(src_cell.shape.height / 2)
                    ],
                    [
                        dst_cell.shape.x + int(dst_cell.shape.width / 2),
                        dst_cell.shape.y + dst_cell.shape.height + edge_end_gap_size
                    ]
                ]


def calc_text(grid):
    for cell in grid.cells:
        shape = getattr(cell.node, 'shape', DEFAULT_SHAPE)
        text_align = shape.get('text-align')
        font_size = shape.get('font-size', 12)
        if text_align == 'middle':
            if cell.node.task_title:
                title = force_str(cell.node.task_title)
            else:
                title = ' '.join(cell.node.name.capitalize().split('_'))
            segments = wrap(title, 20)
            block_height = max(len(segments) - 1, 0) * font_size * 1.2

            x_start = int(cell.shape.width / 2)
            y_start = int((cell.shape.height - block_height) / 2)

            for n, segment in enumerate(segments):
                cell.shape.text.append(
                    (
                        segment, 'align-middle', font_size,
                        x_start,
                        y_start + (n * font_size * 1.2)
                    )
                )


def calc_cell_status(flow_class, grid, process_pk):
    tasks = flow_class.task_class._default_manager.filter(
        process_id=process_pk,
        process__flow_class=flow_class
    ).order_by(
        'flow_task', 'created'
    )
    tasks_group = dict(
        (k, list(v))
        for k, v in groupby(tasks, lambda task: task.flow_task)
    )
    for flow_task in grid.nodes:
        for task in tasks_group.get(flow_task, []):
            if grid[flow_task].status is not None and task.status == STATUS.NEW:
                pass  # don't touch new tasks
            else:
                grid[flow_task].status = task.status


def grid_to_svg(grid):
    svg_template = get_template('viewflow/workflow/graph.svg')
    return svg_template.render({
        'grid': grid,
        'cells': grid.cells,
        'edges': grid.edges})


def grid_to_bpmn(grid):
    svg_template = get_template('viewflow/workflow/graph.bpmn')
    return svg_template.render({
        'grid': grid,
        'cells': grid.cells,
        'edges': grid.edges})



"""Base FSM declarations."""

# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.

from __future__ import annotations

import inspect
from typing import Any, Dict, Mapping, Iterable, List, Type, Optional
from viewflow.this_object import ThisObject
from viewflow.utils import MARKER
from .typing import (
    UserModel,
    Condition,
    StateTransitions,
    Permission,
    StateValue,
    TransitionFunction,
)


class TransitionNotAllowed(Exception):
    """Raised when a transition is not allowed."""


class Transition(object):
    """
    A state transition definition.
    """

    def __init__(
        self,
        func: TransitionFunction,
        source: StateValue,
        target: Optional[StateValue],
        label: Optional[str] = None,
        conditions: Optional[List[Condition]] = None,
        permission: Optional[Permission] = None,
    ):  # noqa D102
        self.func = func
        self.source = source
        self.target = target
        self._label = label
        self.permission = permission
        self.conditions = conditions if conditions else []

    def __repr__(self) -> str:
        return f"<Transition({self.label} {self.source} -> {self.target}) object at {id(self)}>"

    def __str__(self) -> str:
        return f"{self.label} Transition"

    @property
    def label(self) -> str:
        """Return the human-readable label for the transition."""
        if self._label:
            return self._label
        else:
            try:
                return self.func.label  # type: ignore
            except AttributeError:
                return self.func.__name__.title()

    @property
    def slug(self) -> str:
        """Return the slugified version of the transition function name."""
        return self.func.__name__

    def conditions_met(self, instance: object) -> bool:
        """Check if all associated conditions are met for the transition."""
        conditions = [
            condition.resolve(instance.__class__)
            if isinstance(condition, ThisObject)
            else condition
            for condition in self.conditions
        ]
        return all(map(lambda condition: condition(instance), conditions))

    def has_perm(self, instance: object, user: UserModel) -> bool:
        """Check if the user has the required permission to execute the transition."""
        if self.permission is None:
            return False
        elif callable(self.permission):
            return self.permission(instance, user)
        elif isinstance(self.permission, ThisObject):
            permission = self.permission.resolve(instance)
            return permission(user)  # type: ignore
        else:
            raise ValueError(f"Unknown permission type {type(self.permission)}")


class TransitionMethod(object):
    """Unbound transition method wrapper.

    Provides shortcut to enumerate all method transitions, ex::

        Review.publish.get_transitions()
    """

    do_not_call_in_templates = True

    def __init__(
        self,
        state: State,
        func: TransitionFunction,
        descriptor: TransitionDescriptor,
        owner: Type[object],
    ):
        self._state = state
        self._func = func
        self._descriptor = descriptor
        self._owner = owner

        self.__doc__ = func.__doc__

    def get_transitions(self) -> Iterable[Transition]:
        return self._descriptor.get_transitions()

    @property
    def slug(self) -> str:
        """Transition name."""
        return self._func.__name__


class TransitionBoundMethod(object):
    """Instance method wrapper that performs the transition."""

    do_not_call_in_templates = True

    class Wrapper(object):
        """Wrapper context object, to simplify __call__ method debug"""

        def __init__(self, parent: "TransitionBoundMethod", kwargs: Mapping[str, Any]):
            self.parent = parent
            self.caller_kwargs = kwargs
            self.initial_state: StateValue
            self.target_state: StateValue

        def __enter__(self) -> None:
            self.initial_state = self.parent._state.get(self.parent._instance)
            transition = self.parent._descriptor.get_transition(self.initial_state)

            if transition is None:
                raise TransitionNotAllowed(
                    f'{self.parent.label} :: no transition from "{self.initial_state}"'
                )

            if not transition.conditions_met(self.parent._instance):
                raise TransitionNotAllowed(
                    f" '{transition.label}' transition conditions have not been met"
                )

            self.target_state = transition.target
            if self.target_state:
                self.parent._state.set(self.parent._instance, self.target_state)

        def __exit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Any,
        ) -> None:
            if exc_type is not None:
                self.parent._state.set(self.parent._instance, self.initial_state)
            else:
                self.parent._state.transition_succeed(
                    self.parent._instance,
                    self.parent,
                    self.initial_state,
                    self.target_state,
                    **self.caller_kwargs,
                )

    def __init__(
        self,
        state: State,
        func: TransitionFunction,
        descriptor: TransitionDescriptor,
        instance: object,
    ):
        self._state = state
        self._func = func
        self._descriptor = descriptor
        self._instance = instance

    def original(self, *args: Any, **kwargs: Any) -> Any:
        """Call the unwrapped class method."""
        return self._func(self._instance, *args, **kwargs)

    def can_proceed(self, check_conditions: bool = True) -> bool:
        """Check is transition available."""
        current_state = self._state.get(self._instance)
        transition = self._descriptor.get_transition(current_state)
        if transition and check_conditions:
            return transition.conditions_met(self._instance)
        return False

    def has_perm(self, user: UserModel) -> bool:
        current_state = self._state.get(self._instance)
        transition = self._descriptor.get_transition(current_state)
        if transition:
            return transition.has_perm(self._instance, user)
        return False

    @property
    def label(self) -> str:
        """Transition human-readable label."""
        current_state = self._state.get(self._instance)
        transition = self._descriptor.get_transition(current_state)
        if transition:
            return transition.label
        else:
            try:
                return self._func.label  # type: ignore
            except AttributeError:
                return self._func.__name__.title()

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        with TransitionBoundMethod.Wrapper(self, kwargs=kwargs):
            return self._func(self._instance, *args, **kwargs)

    def get_transitions(self) -> Iterable[Transition]:
        return self._descriptor.get_transitions()


class TransitionDescriptor(object):
    """Base transition definition descriptor."""

    do_not_call_in_templates = True

    def __init__(self, state: StateValue, func: TransitionFunction):  # noqa D102
        self._state = state
        self._func = func
        self._transitions: Dict[StateValue, Transition] = {}

    def __get__(
        self, instance: object, owner: Optional[Type[object]] = None
    ) -> TransitionMethod | TransitionBoundMethod:
        if instance:
            return TransitionBoundMethod(self._state, self._func, self, instance)
        else:
            assert owner is not None  # make mypy happy
            return TransitionMethod(self._state, self._func, self, owner)

    def add_transition(self, transition: Transition) -> None:
        self._transitions[transition.source] = transition

    def get_transitions(self) -> Iterable[Transition]:
        """List of all transitions."""
        return self._transitions.values()

    def get_transition(self, source_state: StateValue) -> Optional[Transition]:
        """Get a transition of a source_state.

        Returns None if there is no outgoing transitions.
        """
        transition = self._transitions.get(source_state, None)
        if transition is None:
            transition = self._transitions.get(State.ANY, None)
        return transition


class SuperTransitionDescriptor(object):
    do_not_call_in_templates = True

    def __init__(self, state: State, func: TransitionFunction):  # noqa D102
        self._state = state
        self._func = func

    def __get__(
        self, instance: object, owner: Optional[Type[object]] = None
    ) -> TransitionBoundMethod | TransitionMethod:
        if instance:
            return TransitionBoundMethod(
                self._state,
                self._func,
                self.get_descriptor(instance.__class__),
                instance,
            )
        else:
            assert owner is not None  # make mypy happy
            return TransitionMethod(
                self._state, self._func, self.get_descriptor(owner), owner
            )

    def get_descriptor(self, owner: Type[object]) -> TransitionDescriptor:
        """Lookup for the transition descriptor in the base classes."""
        for cls in owner.__mro__[1:]:
            if hasattr(cls, self._func.__name__):
                super_method = getattr(cls, self._func.__name__)
                if isinstance(super_method, TransitionMethod):
                    break
        else:
            raise ValueError("Base transition not found")

        return super_method._descriptor


class StateDescriptor(object):
    """Class-bound value for a state descriptor.

    Provides shortcut to enumerate all class transitions, ex::

        Review.state.get_transitions()
    """

    def __init__(self, state: "State", owner: type):
        self._state = state
        self._owner = owner

    def __getattr__(self, attr: str) -> Any:
        return getattr(self._state, attr)

    def get_transitions(self) -> StateTransitions:
        propname = "__fsm_{}_transitions".format(self._state.propname)
        transitions: StateTransitions = self._owner.__dict__.get(propname, None)
        if transitions is None:
            transitions = {}

            methods = inspect.getmembers(
                self._owner, lambda attr: isinstance(attr, TransitionMethod)
            )
            transitions = {method: method.get_transitions() for _, method in methods}
            setattr(self._owner, propname, transitions)

        return transitions

    def get_outgoing_transitions(self, state: State) -> List[Transition]:
        return [
            transition
            for transitions in self.get_transitions().values()
            for transition in transitions
            if transition.source == state
            or (transition.source == State.ANY and transition.target != state)
        ]

    def get_available_transitions(self, flow, state: State, user):
        return [
            transition
            for transition in self.get_outgoing_transitions(state)
            if transition.conditions_met(flow)
            if transition.has_perm(flow, user)
        ]


class State(object):
    """State slot field."""

    ANY = MARKER("ANY")

    def __init__(self, states: Any, default: StateValue = None):
        self._default = default
        self._setter = None
        self._getter = None
        self._on_success = None

    def __get__(self, instance: object, owner: Optional[Type[object]] = None) -> Any:
        if instance:
            return self.get(instance)
        assert owner is not None  # make mypy happy
        return StateDescriptor(self, owner)

    def __set__(self, instance: object, value: StateValue) -> None:
        raise AttributeError("Direct state modification is not allowed")

    def get(self, instance: object) -> Any:
        """Get the state from the underline class instance."""
        if self._getter:
            value = self._getter(instance)
            if self._default:
                return value if value else self._default
            else:
                return value
        return getattr(instance, self.propname, self._default)

    def set(self, instance: object, value: StateValue) -> None:
        """Get the state of the underline class instance."""
        if self._setter:
            self._setter(instance, value)
        else:
            setattr(instance, self.propname, value)

    def transition_succeed(
        self,
        instance: object,
        descriptor: TransitionBoundMethod,
        source: State,
        target: State,
        **kwargs: Any,
    ) -> None:
        if self._on_success:
            self._on_success(instance, descriptor, source, target, **kwargs)

    @property
    def propname(self) -> str:
        """State storage attribute."""
        return "__fsm{}".format(id(self))

    def transition(
        self,
        source: StateValue,
        target: Optional[StateValue] = None,
        label: Optional[str] = None,
        conditions: Optional[List[Condition]] = None,
        permission: Optional[Permission] = None,
    ) -> Any:
        """Transition method decorator."""

        def _wrapper(func: Any) -> Any:
            if isinstance(func, TransitionDescriptor):
                descriptor = func
            else:
                descriptor = TransitionDescriptor(self, func)

            source_list = source
            if not isinstance(source, (list, tuple, set)):
                source_list = [source]

            for src in source_list:
                transition = Transition(
                    func=descriptor._func,
                    source=src,
                    target=target,
                    label=label,
                    conditions=conditions,
                    permission=permission,
                )
                descriptor.add_transition(transition)

            return descriptor

        return _wrapper

    def super(self) -> Any:
        def _wrapper(func: Any) -> Any:
            return SuperTransitionDescriptor(self, func)

        return _wrapper

    def setter(self) -> Any:
        def _wrapper(func: Any) -> Any:
            self._setter = func
            return func

        return _wrapper

    def getter(self) -> Any:
        def _wrapper(func: Any) -> Any:
            self._getter = func
            return func

        return _wrapper

    def on_success(self) -> Any:
        def _wrapper(func: Any) -> Any:
            self._on_success = func
            return func

        return _wrapper

    class CONDITION(object):
        """Boolean-like object to return value accompanied with a message from fsm conditions."""

        def __init__(self, is_true: bool, unmet: str = ""):
            self.is_true = is_true
            self.unmet = unmet

        @property
        def message(self) -> str:
            return self.unmet if not self.is_true else ""

        def __bool__(self) -> bool:
            return self.is_true



from django.views.generic import RedirectView
from django.urls import path
from django.utils.http import url_has_allowed_host_and_scheme as is_safe_url
from django.utils.translation import gettext_lazy as _

from viewflow import viewprop
from viewflow.utils import DEFAULT
from viewflow.urls import Application, AppMenuMixin, Viewset, ViewsetMeta
from ..status import STATUS
from ..models import Task
from . import views


class BaseFlowViewsMixin(metaclass=ViewsetMeta):
    """Common Views for Flow and FlowApp viewsets"""

    def __init__(self, flow_class, **kwargs):
        super().__init__(**kwargs)
        self._flow_class = flow_class

    def filter_kwargs(self, view_class, **kwargs):
        return super().filter_kwargs(
            view_class, **{"flow_class": self._flow_class, "viewset": self, **kwargs}
        )

    """
    Permissions
    """

    def has_view_permission(self, user, obj=None):
        return self._flow_class.instance.has_view_permission(user)

    """
    Dashboard
    """

    dashboard_view_class = views.DashboardView

    def get_dashboard_view_kwargs(self, **kwargs):
        return self.filter_kwargs(self.dashboard_view_class, **kwargs)

    @viewprop
    def dashboard_view(self):
        return self.dashboard_view_class.as_view(**self.get_dashboard_view_kwargs())

    @property
    def index_path(self):
        return path("", self.dashboard_view, name="index")

    """
    Process list
    """

    process_list_view_class = views.DashboardProcessListView

    def get_process_list_view_kwargs(self, **kwargs):
        return self.filter_kwargs(self.process_list_view_class, **kwargs)

    @viewprop
    def process_list_view(self):
        return self.process_list_view_class.as_view(
            **self.get_process_list_view_kwargs()
        )

    @property
    def process_list_path(self):
        return path("flows/", self.process_list_view, name="process_list")

    """
    Process detail
    """
    process_detail_view_class = views.DetailProcessView

    def get_process_detail_view_kwargs(self, **kwargs):
        return self.filter_kwargs(self.process_detail_view_class, **kwargs)

    @viewprop
    def process_detail_view(self):
        return self.process_detail_view_class.as_view(
            **self.get_process_detail_view_kwargs()
        )

    @property
    def process_detail_path(self):
        return path(
            "<int:process_pk>/", self.process_detail_view, name="process_detail"
        )

    """
    Process cancel
    """
    process_cancel_view_class = views.CancelProcessView

    def get_process_cancel_view_kwargs(self, **kwargs):
        return self.filter_kwargs(self.process_cancel_view_class, **kwargs)

    @viewprop
    def process_cancel_view(self):
        return self.process_cancel_view_class.as_view(
            **self.get_process_cancel_view_kwargs()
        )

    @property
    def process_cancel_path(self):
        return path(
            "<int:process_pk>/cancel/", self.process_cancel_view, name="process_cancel"
        )

    """
    Task list
    """

    task_list_view_class = views.DashboardTaskListView

    def get_task_list_view_kwargs(self, **kwargs):
        return self.filter_kwargs(self.task_list_view_class, **kwargs)

    @viewprop
    def task_list_view(self):
        return self.task_list_view_class.as_view(**self.get_task_list_view_kwargs())

    @property
    def task_list_path(self):
        return path("tasks/", self.task_list_view, name="task_list")

    """
    Chart View
    """

    chart_view_class = views.FlowChartView

    def get_chart_view_kwargs(self, **kwargs):
        return self.filter_kwargs(self.chart_view_class, **kwargs)

    @viewprop
    def chart_view(self):
        return self.chart_view_class.as_view(**self.get_chart_view_kwargs())

    @property
    def chart_path(self):
        return path("chart/", self.chart_view, name="chart")

    @property
    def flow_chart_path(self):
        return path("<int:process_pk>/chart/", self.chart_view, name="process_chart")

    """
    Utils
    """

    def get_success_url(self, request):
        if not hasattr(request, "activation"):
            return self.reverse("index")

        if "_continue" in request.POST:
            manager = self._flow_class.task_class._default_manager

            next_user_task = request.activation.task
            if next_user_task.status == STATUS.DONE:
                next_user_task = manager.next_user_task(
                    request.activation.process,
                    request.user,
                )

            if next_user_task:
                return next_user_task.flow_task.reverse(
                    "index", args=[next_user_task.process_id, next_user_task.pk]
                )

        if "back" in request.GET:
            back_url = request.GET["back"]
            if not is_safe_url(url=back_url, allowed_hosts={request.get_host()}):
                back_url = "/"
            return back_url

        if hasattr(request, "session") and "vf-pin-location" in request.session:
            back_url = request.session.get("vf-pin-location", "")
            if not is_safe_url(url=back_url, allowed_hosts={request.get_host()}):
                back_url = "/"
            return back_url

        return self.reverse("process_detail", args=[request.activation.process.pk])


class BulkActionsViewsMixin(metaclass=ViewsetMeta):
    """
    Bulk Assign Tasks
    """

    tasks_assign_view_class = views.BulkAssignTasksActionView

    def get_tasks_assign_view_kwargs(self, **kwargs):
        return self.filter_kwargs(self.tasks_assign_view_class, **kwargs)

    @viewprop
    def tasks_assign_view(self):
        return self.tasks_assign_view_class.as_view(
            **self.get_tasks_assign_view_kwargs()
        )

    @property
    def tasks_assign_path(self):
        return path("queue/assign/", self.tasks_assign_view, name="tasks_assign")

    """
    Bulk Unassign Tasks
    """
    tasks_unassign_view_class = views.BulkUnassignTasksActionView

    def get_tasks_unassign_view_kwargs(self, **kwargs):
        return self.filter_kwargs(self.tasks_unassign_view_class, **kwargs)

    @viewprop
    def tasks_unassign_view(self):
        return self.tasks_unassign_view_class.as_view(
            **self.get_tasks_unassign_view_kwargs()
        )

    @property
    def tasks_unassign_path(self):
        return path("inbox/unassign/", self.tasks_unassign_view, name="tasks_unassign")


class FlowViewset(BaseFlowViewsMixin, AppMenuMixin, Viewset):
    """
    Basic flow viewset to include single flow into url patterns

    Usage:

    .. code-block:: python

        urlpatterns = [path("my_flow/", FlowViewset(MyFlow).urls)]

    """

    def _get_urls(self):
        own_patterns = super()._get_urls()
        flow_patterns, _, _ = self._flow_class.instance.urls
        self._flow_class.parent = self
        self._flow_class.app_name = None

        return own_patterns + flow_patterns

    @viewprop
    def app_name(self):
        # assert (
        #     self._flow_class.instance.app_name
        # ), "Flow can't be connected to urlpatterns more than one time"
        return self._flow_class.instance.app_name

    @viewprop
    def title(self):
        return self._flow_class.process_title

    def _get_resolver_extra(self):
        return {"flow_viewset": self, **super()._get_resolver_extra()}


class FlowAppViewset(BaseFlowViewsMixin, BulkActionsViewsMixin, Application):
    """
    Viewset includes flow as an separate App into Site.

    `Cookbook sample <https://github.com/viewflow/cookbook/blob/main/workflow101/config/urls.py>`_

    Usage:

    .. code-block:: python

        site = Site(
            viewsets=[
                FlowAppViewset(
                    ShipmentFlow,
                    icon="local_shipping",
                    viewsets=[
                        ShipmentCRUDViewset()
                    ]
            ]
        )

        urlpatterns = [path("", site.urls)]
    """

    menu_template_name = "viewflow/workflow/flow_menu.html"
    base_template_name = "viewflow/workflow/base_page.html"

    """
    Inbox
    """

    inbox_view_class = views.FlowInboxListView

    def get_inbox_view_kwargs(self, **kwargs):
        return self.filter_kwargs(self.inbox_view_class, **kwargs)

    @viewprop
    def inbox_view(self):
        return self.inbox_view_class.as_view(**self.get_inbox_view_kwargs())

    @property
    def inbox_path(self):
        return path("inbox/", self.inbox_view, name="inbox")

    """
    Queue
    """

    queue_view_class = views.FlowQueueListView

    def get_queue_view_kwargs(self, **kwargs):
        return self.filter_kwargs(self.queue_view_class, **kwargs)

    @viewprop
    def queue_view(self):
        return self.queue_view_class.as_view(**self.get_queue_view_kwargs())

    @property
    def queue_path(self):
        return path("queue/", self.queue_view, name="queue")

    """
    Archive
    """

    archive_view_class = views.FlowArchiveListView

    def get_archive_view_kwargs(self, **kwargs):
        return self.filter_kwargs(self.archive_view_class, **kwargs)

    @viewprop
    def archive_view(self):
        return self.archive_view_class.as_view(**self.get_archive_view_kwargs())

    @property
    def archive_path(self):
        return path("archive/", self.archive_view, name="archive")

    """
    Other staff
    """

    def _get_urls(self):
        own_patterns = super()._get_urls()
        flow_patterns, _, _ = self._flow_class.instance.urls
        self._flow_class.parent = self
        self._flow_class.app_name = None

        return own_patterns + flow_patterns

    def _get_resolver_extra(self):
        return {
            "app": self,
            "viewset": self,
            "flow_viewset": self,
            "flow": self._flow_class.instance,
        }

    @viewprop
    def app_name(self):
        return self._flow_class.instance.app_name

    @viewprop
    def title(self):
        return self._flow_class.process_title

    def get_context_data(self, request):
        manager = self._flow_class.task_class._default_manager

        inbox = manager.inbox([self._flow_class], request.user.id)

        queue = manager.queue([self._flow_class], request.user.id)

        return {
            "user_inbox": inbox,
            "user_queue": queue,
        }


class _IndexRedirectView(RedirectView):
    viewset = None

    def get_redirect_url(self, *args, **kwargs):
        if self.viewset:
            redirect = None

            for flow_class in self.viewset.parent.flow_classes:
                if flow_class.instance.has_view_permission(self.request.user):
                    redirect = flow_class.instance.reverse("index")
                    break

            if redirect is None:
                raise ValueError(
                    "Can't determine index url. Please add an explicit "
                    "`index_path = path('', RedirectView(url='...'), name='index')`"
                    " declaration for the viewset"
                )
            return redirect
        return super().get_redirect_url(*args, **kwargs)


class NestedFlowsApp(AppMenuMixin, Application):
    app_name = "flows"
    title = _("Processes")
    icon = "notes"

    """
    Permissions
    """

    def has_view_permission(self, user, obj=None):
        return self.parent.has_view_permission(user, obj=obj)

    @property
    def index_path(self):
        return path("", _IndexRedirectView.as_view(viewset=self), name="index")


class WorkflowAppViewset(BulkActionsViewsMixin, Application):
    """
    Viewset includes multiples flow with common Inbox/Queue/Archive views as an
    separate App into Site.

    `Life demo
    <https://demo.viewflow.io/workflow/inbox/>`_

    Usage:

    .. code-block:: python

        site = Site(
            viewsets=[
                WorkflowAppViewset(
                    flow_viewsets=[
                        FlowViewset(HelloWorldFlow, icon="assignment"),
                        FlowViewset(ShipmentFlow, icon="local_shipping"),
                        FlowViewset(DynamicSplitFlow, icon="tune"),
                    ]
            ]
        )

        urlpatterns = [path("", site.urls)]
    """

    app_name = "workflow"
    icon = "assignment"
    menu_template_name = "viewflow/workflow/workflow_menu.html"
    base_template_name = "viewflow/workflow/base_page.html"

    def __init__(self, flow_viewsets, **kwargs):
        self.flow_classes = [viewset._flow_class for viewset in flow_viewsets]

        viewsets = kwargs.get("viewsets", [])
        viewsets.append(NestedFlowsApp(viewsets=flow_viewsets))
        kwargs["viewsets"] = viewsets
        super().__init__(**kwargs)

    def _get_resolver_extra(self):
        return {
            "app": self,
            "viewset": self,
        }

    def filter_kwargs(self, view_class, **kwargs):
        return super().filter_kwargs(
            view_class, **{"flow_classes": self.flow_classes, "viewset": self, **kwargs}
        )

    """
    Permissions
    """

    def has_view_permission(self, user, obj=None):
        return any(
            flow_class.instance.has_view_permission(user, obj=obj)
            for flow_class in self.flow_classes
        )

    """
    Inbox
    """

    inbox_view_class = views.WorkflowInboxListView

    def get_inbox_view_kwargs(self, **kwargs):
        return self.filter_kwargs(self.inbox_view_class, **kwargs)

    @viewprop
    def inbox_view(self):
        return self.inbox_view_class.as_view(**self.get_inbox_view_kwargs())

    @property
    def inbox_path(self):
        return path("inbox/", self.inbox_view, name="inbox")

    """
    Queue
    """

    queue_view_class = views.WorkflowQueueListView

    def get_queue_view_kwargs(self, **kwargs):
        return self.filter_kwargs(self.queue_view_class, **kwargs)

    @viewprop
    def queue_view(self):
        return self.queue_view_class.as_view(**self.get_queue_view_kwargs())

    @property
    def queue_path(self):
        return path("queue/", self.queue_view, name="queue")

    """
    Archive
    """

    archive_view_class = views.WorkflowArchiveListView

    def get_archive_view_kwargs(self, **kwargs):
        return self.filter_kwargs(self.archive_view_class, **kwargs)

    @viewprop
    def archive_view(self):
        return self.archive_view_class.as_view(**self.get_archive_view_kwargs())

    @property
    def archive_path(self):
        return path("archive/", self.archive_view, name="archive")

    def get_context_data(self, request):
        inbox = Task.objects.inbox(self.flow_classes, request.user)

        queue = Task.objects.queue(self.flow_classes, request.user)

        return {
            "user_inbox": inbox,
            "user_queue": queue,
        }



from django_filters.filterset import filterset_factory


class FilterableViewMixin(object):
    """
    A mixin that provides a way to show and handle a FilterSet in a request.
    """

    filterset_class = None
    filter_fields = None
    strict_filter = False

    def get_filterset_class(self):
        """
        Returns the filterset class to use in this view
        """
        if self.filterset_class:
            return self.filterset_class
        elif self.filter_fields:
            return filterset_factory(model=self.model, fields=self.filter_fields)

    def get_filterset_kwargs(self, filterset_class):
        """
        Returns the keyword arguments for instantiating the filterset.
        """
        kwargs = {}

        if self.viewset is not None and hasattr(self.viewset, "get_filterset_kwargs"):
            kwargs = self.viewset.get_filterset_kwargs(self.request)

        return {
            **kwargs,
            "data": self.request.GET or None,
            "request": self.request,
        }

    def get_filterset(self, filterset_class, queryset):
        """
        Returns an instance of the filterset to be used in this view.
        """
        kwargs = self.get_filterset_kwargs(filterset_class)
        return filterset_class(queryset=queryset, **kwargs)

    def is_strict_filter(self):
        return self.strict_filter

    def get_queryset(self):
        queryset = super().get_queryset()
        self.filterset, filterset_class = None, self.get_filterset_class()
        if filterset_class is not None:
            self.filterset = self.get_filterset(filterset_class, queryset=queryset)
            if self.filterset.is_valid() or not self.is_strict_filter():
                queryset = self.filterset.qs
            else:
                queryset = self.filterset.queryset.none()
        return queryset



from django.apps import AppConfig

# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.


class ViewflowAdminConfig(AppConfig):
    """Default application config."""

    name = "viewflow.contrib.admin"
    label = "viewflow_admin"



from functools import lru_cache

from django.db import models
from django.utils.module_loading import import_string

from viewflow.utils import get_app_package, get_containing_app_data
from viewflow.workflow.exceptions import FlowRuntimeError
from viewflow.workflow.token import Token


@lru_cache(maxsize=None)
def import_flow_by_ref(flow_strref):
    """Return flow class by flow string reference."""
    app_label, flow_path = flow_strref.split("/")
    # TODO Raise if imported class is not Flow
    return import_string("{}.{}".format(get_app_package(app_label), flow_path))


@lru_cache(maxsize=None)
def get_flow_ref(flow_class):
    """Convert flow class to string reference."""
    module = "{}.{}".format(flow_class.__module__, flow_class.__name__)
    app_label, app_package = get_containing_app_data(module)
    if app_label is None:
        raise FlowRuntimeError(
            "No application found for {}. Check your INSTALLED_APPS setting".format(
                module
            )
        )

    subpath = module[len(app_package) + 1 :]
    return "{}/{}".format(app_label, subpath)


@lru_cache(maxsize=None)
def import_task_by_ref(task_strref):
    """Return flow task by reference like `app_label/path.to.Flowclass.task_name`."""
    app_label, flow_path = task_strref.split("/")
    flow_path, task_name = flow_path.rsplit(".", 1)
    flow_class = import_string("{}.{}".format(get_app_package(app_label), flow_path))
    # TODO Raise if imported class is not Node
    return flow_class.instance.node(task_name)


@lru_cache(maxsize=None)
def get_task_ref(flow_task):
    """Convert task to the string reference suitable to store in the db."""
    # fmt: off
    if flow_task is None or flow_task.flow_class is None:
        return None
    module = flow_task.flow_class.__module__
    app_label, app_package = get_containing_app_data(module)
    if app_label is None:
        raise FlowRuntimeError(
            "No application found for {}. Check your INSTALLED_APPS setting".format(
                module
            )
        )

    subpath = module[len(app_package) + 1 :]

    return "{}/{}.{}.{}".format(
        app_label, subpath, flow_task.flow_class.__name__, flow_task.name
    )


class FlowReferenceField(models.CharField):
    description = """Flow class reference field,
    stores flow as app_label/flows.FlowName> to
    avoid possible collisions with app name changes"""

    def __init__(self, *args, **kwargs):  # noqa D1o2
        kwargs.setdefault("max_length", 250)
        super(FlowReferenceField, self).__init__(*args, **kwargs)

    def to_python(self, value):  # noqa D102
        if value:
            return import_flow_by_ref(value)
        return value

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        try:
            return import_flow_by_ref(value)
        except LookupError:
            return None

    def get_prep_value(self, value):  # noqa D1o2
        if value and not isinstance(value, str):
            return get_flow_ref(value)
        return value

    def value_to_string(self, obj):  # noqa D1o2
        value = super(FlowReferenceField, self).value_from_object(obj)
        return self.get_prep_value(value)


class TaskReferenceField(models.CharField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("max_length", 255)
        super(TaskReferenceField, self).__init__(*args, **kwargs)

    def to_python(self, value):
        if value:
            return import_task_by_ref(value)  # TODO Raise  ValidationError
        return value

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        return import_task_by_ref(value)

    def get_prep_value(self, value):  # noqa D102
        if value and not isinstance(value, str):
            return get_task_ref(value)
        return value

    def value_to_string(self, obj):  # noqa D102
        value = super(TaskReferenceField, self).value_from_object(obj)
        return self.get_prep_value(value)


class TokenField(models.CharField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("max_length", 150)
        super(TokenField, self).__init__(*args, **kwargs)

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        return Token(value)

    def to_python(self, value):
        return Token(value)

    def get_prep_value(self, value):
        return value.token



from rest_framework.pagination import PageNumberPagination


def get_client_ip(request):
    """Получить ip клиента из request"""
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


class CurrentUserIdDefault:
    """Получить user_id из request для сериалайзера"""
    requires_context = True

    def __call__(self, serializer_field):
        return serializer_field.context['request'].user.id

    def __repr__(self):
        return '%s()' % self.__class__.__name__


class ClientIpDefault:
    """Получить ip клиента из request для сериалайзера"""
    requires_context = True

    def __call__(self, serializer_field):
        return get_client_ip(serializer_field.context['request'])

    def __repr__(self):
        return '%s()' % self.__class__.__name__


class PageSizePagination(PageNumberPagination):
    page_size_query_param = 'page_size'



# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.
import copy
import types
import warnings
from collections import namedtuple, OrderedDict

from django.views.generic import RedirectView
from django.urls import URLPattern, URLResolver, include, path, reverse
from django.urls.resolvers import RoutePattern

from viewflow.utils import (
    camel_case_to_underscore,
    strip_suffixes,
    list_path_components,
    DEFAULT,
)


class _UrlName(str):
    """
    Dump str wrapper.

    Just to keep a reference over django url resolve calling
    hierarchy.
    """

    def __init__(self, value):
        str.__init__(value)
        self.extra = {}


class _URLResolver(URLResolver):
    def __init__(self, *args, **kwargs):
        self.extra = kwargs.pop("extra", {})
        super(_URLResolver, self).__init__(*args, **kwargs)

    def resolve(self, *args, **kwargs):
        result = super(_URLResolver, self).resolve(*args, **kwargs)
        if not isinstance(result.url_name, _UrlName):
            result.url_name = _UrlName(result.url_name)

        extra = {}
        extra.update(self.extra)
        extra.update(result.url_name.extra)
        result.url_name.extra = extra
        return result


Route = namedtuple("Route", ["prefix", "viewset"])


def route(prefix, viewset):
    """A viewset url"""
    if not isinstance(viewset, BaseViewset):
        raise ValueError(
            "route(...) second argument should be a viewset instance, got {} instead".format(
                viewset
            )
        )
    return Route(prefix, viewset)


class BaseViewset(object):
    """
    Class-based django routing configuration
    """

    app_name = None
    namespace = None
    parent_namespace = None
    extra_kwargs = None

    def __init__(self):
        super().__init__()
        self._parent = None

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, value):
        if self._parent is not None:
            warnings.warn(
                f"Viewset {self.__class__.__name__} parent could be set only once",
                Warning,
            )
        if self.parent_namespace is not None:
            warnings.warn(
                f"Viewset {self.__class__.__name__} already has explicit parent namespace",
                Warning,
            )
        self._parent = value

    @property
    def urls(self):
        raise NotImplementedError("Subclass should override this")

    def reverse(self, viewname, args=None, kwargs=None, current_app=None):
        """Get view url."""
        view_namespace = ""

        current_viewset = self
        while current_viewset:
            namespace = current_viewset.namespace or current_viewset.app_name
            if namespace:
                view_namespace = "{}:{}".format(namespace, view_namespace)
            if current_viewset.parent is None and current_viewset.parent_namespace:
                view_namespace = "{}:{}".format(
                    current_viewset.parent_namespace, view_namespace
                )
            current_viewset = current_viewset.parent

        if view_namespace:
            viewname = view_namespace + viewname

        return reverse(viewname, args=args, kwargs=kwargs, current_app=current_app)

    def has_view_permission(self, user, obj=None):
        return True


class ViewsetMeta(type):
    """Collect urls declared on the based classes."""

    def __new__(mcs, name, bases, attrs):
        current_patterns = []
        for key, value in list(attrs.items()):
            if not key.endswith("_path") or key.startswith("get_"):
                continue
            current_patterns.append((key, value))

        attrs["declared_patterns"] = OrderedDict()  # current_patterns removed

        new_class = super().__new__(mcs, name, bases, attrs)

        # Walk through the MRO.
        declared_patterns = OrderedDict()
        for base in reversed(new_class.__mro__):
            # Collect fields from base class.
            if hasattr(base, "declared_patterns"):
                declared_patterns.update(base.declared_patterns)

            # Field shadowing.
            for attr, value in base.__dict__.items():
                if value is None and attr in declared_patterns:
                    declared_patterns.pop(attr)

        # place current class items on top of inherited, to allow override first index view in a child class
        # new_class.declared_patterns = declared_patterns
        new_class.declared_patterns = OrderedDict(current_patterns)
        new_class.declared_patterns.update(
            {
                key: value
                for key, value in declared_patterns.items()
                if key not in current_patterns
            }
        )

        return new_class


class Viewset(BaseViewset, metaclass=ViewsetMeta):
    """
    Declarative, class-based django routing configuration

    An Viewset class automatically collect URL patterns from class attributes
    with names ends on `_url` and includes nested Viewset classes from
    attributes with names ends on '_urls'.

    Viewset classes could be inherited, extended, and have overridden attributes.
    """

    viewsets = None
    turbo_disabled = False
    urlpatterns = None

    def __init__(self, **initkwargs):
        super().__init__()
        self._urls_cache = None
        self._children = []

        # customize instance attributes
        for key, value in initkwargs.items():
            if key.startswith("_"):
                raise TypeError(
                    "You tried to pass the private {} name as a "
                    "keyword argument to {}(). Don't do that.".format(
                        key, self.__name__
                    )
                )
            if not hasattr(self.__class__, key):
                raise TypeError(
                    "{}() received an invalid keyword {}. Viewset constructor "
                    "only accepts arguments that are already "
                    "attributes of the class.".format(self.__class__.__name__, key)
                )
            setattr(self, key, value)

        # clone additional routes
        if self.viewsets:
            self.viewsets = [copy.copy(viewset) for viewset in self.viewsets]
            self._children += self.viewsets

        # clone route instances
        for attr_name in self.declared_patterns:
            attr_value = getattr(self.__class__, attr_name)
            if isinstance(attr_value, Route):
                viewset = copy.copy(attr_value.viewset)
                viewset.extra_kwargs = list_path_components(attr_value.prefix)
                setattr(self, attr_name, Route(attr_value.prefix, viewset))
                self._children.append(viewset)

    def _create_url_pattern(self, value):
        if isinstance(value, (URLPattern, URLResolver)):
            return value
        elif isinstance(value, Route):
            value.viewset.parent = self
            patterns, app_name, namespace = value.viewset.urls
            pattern = path(
                value.prefix if value.prefix else "",
                include((patterns, app_name), namespace=namespace),
            )
            return pattern
        else:
            raise ValueError(
                "{} got unknown url entry {}".format(self.__class__.__name__, value)
            )

    def _get_urls(self):
        """
        Collect URLs from the instance attributes.
        """
        urlpatterns = []

        # url overrides
        for url_entry in self.urlpatterns or []:
            urlpatterns.append(self._create_url_pattern(url_entry))

        # class attributes
        for attr_name in self.declared_patterns:
            attr_value = getattr(self, attr_name)
            if isinstance(attr_value, types.FunctionType) or attr_value is None:
                continue
            urlpatterns.append(self._create_url_pattern(attr_value))

        # additional routes
        for viewset in self.viewsets or []:
            if viewset.app_name is None:
                viewset.app_name = camel_case_to_underscore(
                    strip_suffixes(
                        viewset.__class__.__name__,
                        ["App", "Application", "Viewset", "Admin", "Flow"],
                    )
                )
                assert viewset.app_name, "Can't provide auto name for a {}".format(
                    viewset
                )
            urlpatterns.append(
                self._create_url_pattern(route(f"{viewset.app_name}/", viewset))
            )

        return urlpatterns

    def _get_resolver_extra(self):
        return {"viewset": self}

    def filter_kwargs(self, view_class, **kwargs):
        return {
            name: value
            for name, value in kwargs.items()
            if hasattr(view_class, name)
            if value is not DEFAULT
        }

    @property
    def urls(self):
        namespace = self.namespace
        if namespace is None:
            namespace = self.app_name
        if self._urls_cache is None:
            self._urls_cache = self._get_urls()

        pattern = RoutePattern("", is_endpoint=False)
        resolver = _URLResolver(
            pattern, self._urls_cache, extra=self._get_resolver_extra()
        )
        return [resolver], self.app_name, namespace


def _get_index_redirect_url(viewset):
    """
    Return first non-parameterized viewset url.
    """

    def _get_index_url(url_patterns, prefix="./"):
        index_first_patterns = sorted(
            url_patterns, key=lambda pat: getattr(pat, "name", "") != "index"
        )
        for url_pattern in index_first_patterns:
            if isinstance(url_pattern, URLPattern):
                couldbe_index_view = (
                    isinstance(url_pattern.pattern, RoutePattern)
                    and url_pattern.pattern.converters == {}
                    and not (
                        hasattr(url_pattern.callback, "view_class")
                        and url_pattern.callback.view_class == _IndexRedirectView
                    )
                )
                if couldbe_index_view:
                    return prefix + url_pattern.pattern._route
            elif isinstance(url_pattern, URLResolver) and isinstance(
                url_pattern.pattern, RoutePattern
            ):
                return _get_index_url(
                    url_pattern.url_patterns, prefix + url_pattern.pattern._route
                )

    return _get_index_url(viewset.urls[0], "./")


class _IndexRedirectView(RedirectView):
    viewset = None

    def get_redirect_url(self, *args, **kwargs):
        if self.viewset:
            redirect = _get_index_redirect_url(self.viewset)
            if redirect is None:
                raise ValueError(
                    "Can't determine index url. Please add an explicit "
                    "`index_path = path('', RedirectView(url='...'), name='index')`"
                    " declaration for the viewset"
                )
            return redirect
        return super().get_redirect_url(*args, **kwargs)


class IndexViewMixin(metaclass=ViewsetMeta):
    """
    Redirect from / to the first non-parameterized view of the Viewset class.
    """

    @property
    def index_path(self):
        return path("", _IndexRedirectView.as_view(viewset=self), name="index")



from django.utils.timezone import now
from viewflow import this
from ..activation import Activation, leading_tasks_canceled
from ..base import Node
from ..status import STATUS
from ..signals import task_started, task_finished
from . import mixins


class HandleActivation(mixins.NextNodeActivationMixin, Activation):
    @Activation.status.super()
    def activate(self):
        """Do nothing on sync call"""

    @Activation.status.transition(source=STATUS.NEW, target=STATUS.STARTED)
    def run(self, func, **kwargs):
        self.task.started = now()
        task_started.send(sender=self.flow_class, process=self.process, task=self.task)
        return func(self, **kwargs) if func else self.process

    @Activation.status.transition(source=STATUS.STARTED, target=STATUS.DONE)
    def complete(self):
        """Complete task and create next."""
        super().complete.original()

    @Activation.status.transition(source=STATUS.STARTED)
    def execute(self):
        self.complete()
        self.activate_next()

    @Activation.status.transition(
        source=STATUS.DONE, target=STATUS.CANCELED, conditions=[leading_tasks_canceled]
    )
    def undo(self):
        if self.flow_task._undo_func is not None:
            self.flow_task._undo_func(self)
        super().undo.original()

    @Activation.status.transition(source=STATUS.NEW, target=STATUS.CANCELED)
    def cancel(self):
        self.task.finished = now()
        self.task.save()


class Handle(mixins.NextNodeMixin, Node):
    """
    Task to be executed outside of the flow.
    """

    task_type = "FUNCTION"
    activation_class = HandleActivation

    shape = {
        "width": 150,
        "height": 100,
        "text-align": "middle",
        "svg": """
            <rect class="task" width="150" height="100" rx="5" ry="5"/>
            <path class="task-marker"
                  d="m 15,20 c 10,-6 -8,-8 3,-15 l -9,0 c -11,7 7,9 -3,15 z
                     m -7,-12 l 5,0 m -4.5,3 l 4.5,0 m -3,3 l 5,0 m -4,3 l 5,0"/>
        """,
    }

    bpmn_element = "scriptTask"

    def __init__(self, func=None, undo_func=None, **kwargs):
        super().__init__(**kwargs)
        self._func = func
        self._undo_func = undo_func

    def _create_wrapper_function(self, origin_func, task):
        def func(**kwargs):
            with self.flow_class.lock(task.process.pk):
                task.refresh_from_db()
                activation = self.activation_class(task)
                result = activation.run(origin_func, **kwargs)
                activation.execute()
                return result

        return func

    def run(self, task, **kwargs):
        func = this.resolve(self.flow_class.instance, self._func)
        wrapper = self._create_wrapper_function(func, task)
        return wrapper(**kwargs)



from django.contrib import admin
from django.contrib.admin.widgets import AdminTextInputWidget
from django.contrib.auth.admin import UserAdmin, GroupAdmin
from django.contrib.auth.models import Group, User
from django.db.models import ForeignKey, OneToOneField
from rest_framework.authtoken.models import TokenProxy

from . import models


admin.site.unregister(User)
admin.site.unregister(Group)
admin.site.unregister(TokenProxy)


@admin.register(models.Department)
class DepartmentAdmin(admin.ModelAdmin):
    list_display = ('id', 'created', 'name', 'comment')

    def has_delete_permission(self, request, obj=None):
        return False


@admin.register(Group)
class GroupAdmin(GroupAdmin):
    def has_delete_permission(self, request, obj=None):
        return False


@admin.register(User)
class UserAdmin(UserAdmin):
    list_display = ('id', 'username', 'first_name', 'email', 'is_active', 'is_staff', 'is_superuser', 'date_joined')
    search_fields = ('=id', '=username')
    list_filter = ('is_staff', 'is_active', 'is_superuser', 'date_joined')
    readonly_fields = ('is_active', 'is_superuser', 'password', 'user_permissions', 'date_joined', 'last_login',
                       'last_name',
                       )

    def has_delete_permission(self, request, obj=None):
        return False


@admin.register(models.Ext)
class ExtAdmin(admin.ModelAdmin):
    list_display = ('created', 'updated', 'user', 'phonenumber', 'department', 'ip', 'http_referer')
    search_fields = ('=user__id', '=phonenumber')
    ordering = ('-created', )
    formfield_overrides = {
        ForeignKey: {'widget': AdminTextInputWidget},
        OneToOneField: {'widget': AdminTextInputWidget},
    }

    def has_delete_permission(self, request, obj=None):
        return False



# Generated by Django 4.2.11 on 2024-03-29 11:19

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import viewflow.workflow.fields
import viewflow.workflow.token


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Process',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('flow_class', viewflow.workflow.fields.FlowReferenceField(max_length=250, verbose_name='Flow')),
                ('status', models.CharField(choices=[('CANCELED', 'Canceled'), ('DONE', 'Done'), ('NEW', 'New')], default='NEW', max_length=50, verbose_name='Status')),
                ('created', models.DateTimeField(default=django.utils.timezone.now, verbose_name='Created')),
                ('finished', models.DateTimeField(blank=True, null=True, verbose_name='Finished')),
                ('data', models.JSONField(blank=True, default=dict)),
                ('artifact_object_id', models.PositiveIntegerField(blank=True, null=True)),
            ],
            options={
                'verbose_name': 'Process',
                'verbose_name_plural': 'Process list',
                'ordering': ['-created'],
            },
        ),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('flow_task', viewflow.workflow.fields.TaskReferenceField(max_length=255, verbose_name='Task')),
                ('flow_task_type', models.CharField(max_length=50, verbose_name='Type')),
                ('status', models.CharField(choices=[('ASSIGNED', 'Assigned'), ('CANCELED', 'Canceled'), ('DONE', 'Done'), ('ERROR', 'Error'), ('NEW', 'New'), ('REVIVED', 'Revived'), ('SCHEDULED', 'Scheduled'), ('STARTED', 'Started')], db_index=True, default='NEW', max_length=50, verbose_name='Status')),
                ('created', models.DateTimeField(db_index=True, default=django.utils.timezone.now, verbose_name='Created')),
                ('assigned', models.DateTimeField(blank=True, null=True, verbose_name='Assigned')),
                ('started', models.DateTimeField(blank=True, null=True, verbose_name='Started')),
                ('finished', models.DateTimeField(blank=True, null=True, verbose_name='Finished')),
                ('token', viewflow.workflow.fields.TokenField(default=viewflow.workflow.token.Token('start'), max_length=150, verbose_name='Token')),
                ('external_task_id', models.CharField(blank=True, db_index=True, max_length=50, null=True, verbose_name='External Task ID')),
                ('owner', models.PositiveIntegerField(blank=True, null=True, verbose_name='Сотрудник')),
                ('owner_permission', models.CharField(blank=True, max_length=255, null=True, verbose_name='Permission')),
                ('owner_permission_obj_pk', models.CharField(blank=True, max_length=255, null=True)),
                ('data', models.JSONField(blank=True, default=dict)),
                ('artifact_object_id', models.PositiveIntegerField(blank=True, null=True)),
                ('previous', models.ManyToManyField(related_name='leading', to='viewflow.task', verbose_name='Previous')),
                ('process', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='viewflow.process', verbose_name='Process')),
            ],
            options={
                'verbose_name': 'Task',
                'verbose_name_plural': 'Tasks',
                'ordering': ['-created'],
            },
        ),
        migrations.AddField(
            model_name='process',
            name='parent_task',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='viewflow.task'),
        ),
    ]



from django.utils.timezone import now

from viewflow import this
from viewflow.workflow.base import Node, Edge
from viewflow.workflow.exceptions import FlowRuntimeError
from viewflow.workflow.activation import Activation

from . import mixins
from ..status import STATUS


class SwitchActivation(Activation):
    """Switch gateway activation."""

    next_task = None

    @Activation.status.super()
    def activate(self):
        for node, cond in self.flow_task._branches:
            if cond:
                if cond(self):
                    self.next_task = node
                    break
            else:
                self.next_task = node

        if not self.next_task:
            raise FlowRuntimeError(
                "No next task available for {}".format(self.flow_task.name)
            )

    @Activation.status.super()
    def create_next(self):
        yield self.next_task._create(self, self.task.token)


class Switch(Node):
    """
    Gateway that selects one of the outgoing node.

    Activates first node with matched condition.

    Example::

        select_responsible_person = (
            flow.Switch()
            .Case(this.dean_approval, lambda act: a.process.need_dean)
            .Case(this.head_approval, lambda act: a.process.need_head)
            .Default(this.supervisor_approval)
        )

    """

    task_type = "SWITCH"
    activation_class = SwitchActivation

    def __init__(self, **kwargs):  # noqa D102
        super(Switch, self).__init__(**kwargs)
        self._activate_next = []

    def _outgoing(self):
        for next_node, cond in self._activate_next:
            edge_class = "cond_true" if cond else "default"
            yield Edge(src=self, dst=next_node, edge_class=edge_class)

    def _resolve(self, instance):
        super()._resolve(instance)

        next_nodes = []
        for node, condition in self._activate_next:
            node = this.resolve(instance, node)
            condition = this.resolve(instance, condition)
            next_nodes.append((node, condition))
        self._activate_next = next_nodes

    @property
    def _branches(self):
        return self._activate_next

    def Case(self, node, cond=None):
        """Node to activate if condition is True.

        :param cond: Calable[activation] -> bool
        """
        self._activate_next.append((node, cond))
        return self

    def Default(self, node):
        """Last node to activate if no one other succeed."""
        self._activate_next.append((node, None))
        return self



from django.urls import path
from django.utils.functional import cached_property
from viewflow.urls import AppMenuMixin, Viewset
from viewflow.utils import viewprop, DEFAULT

from dash import Dash

from . import views


class Dashboard(AppMenuMixin, Viewset):
    layout = None
    dashboard_template_name = DEFAULT
    turbo_disabled = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._registered_callbacks = []

    @cached_property
    def dash_app(self):
        app = Dash("app", requests_pathname_prefix=self.reverse("index"))
        app.layout = self.layout
        for func, args, kwargs in self._registered_callbacks:
            app.callback(*args, **kwargs)(func)
        return app

    def filter_kwargs(self, view_class, **kwargs):
        return super().filter_kwargs(view_class, **{"viewset": self, **kwargs})

    """
    Dashboard View
    """
    dashboard_view_class = views.DashboardView

    @property
    def index_path(self):
        return path("", self.dashboard_view, name="index")

    @viewprop
    def dashboard_view_kwargs(self):
        return {}

    def get_dashboard_view_kwargs(self, **kwargs):
        view_kwargs = {
            "template_name": self.dashboard_template_name,
            **self.dashboard_view_kwargs,
            **kwargs,
        }
        return self.filter_kwargs(self.dashboard_view_class, **view_kwargs)

    @viewprop
    def dashboard_view(self):
        return self.dashboard_view_class.as_view(**self.get_dashboard_view_kwargs())

    """
    Dash endpoints
    """

    @property
    def layout_path(self):
        return path("_dash-layout/", views.layout_endpoint, {"viewset": self})

    @property
    def dependencies_path(self):
        return path(
            "_dash-dependencies/", views.dependencies_endpoint, {"viewset": self}
        )

    @property
    def update_component_path(self):
        return path(
            "_dash-update-component", views.update_component_endpoint, {"viewset": self}
        )

    """
    Callback
    """

    def callback(self, *args, **kwargs):
        def decorator(func):
            self._registered_callbacks.append((func, args, kwargs))
            return func

        return decorator



from django.contrib import admin

from . import models


class ProlongateFilter(admin.SimpleListFilter):
    title = 'Пролонгированный займ'
    parameter_name = "is_prolongate"

    def lookups(self, request, model_admin):
        return (
            ('yes', 'Да'),
        )

    def queryset(self, request, queryset):
        if self.value() == "yes":
            return queryset.filter(status='OPEN', extended_end_date__isnull=False)
        else:
            return queryset


@admin.register(models.Loan)
class LoanAdmin(admin.ModelAdmin):
    list_display = ('id', 'created', 'status', 'order', 'user_id', 'expiry_date', 'closed_date',
                    'interest_rate', 'period', 'free_period',
                    'body_issue', 'body_paid', 'body_balance',
                    'interests_charged', 'interests_paid', 'interests_balance',
                    'balance',
                    )
    list_filter = ('status', 'created', ProlongateFilter)
    search_fields = ('=id', '=order__id', '=order__user_id')

    @staticmethod
    def user_id(obj):
        return obj.order.user_id

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False
